# HydDown hydrogen/other gas depressurisation
# Copyright (c) 2021 Anders Andreasen
# Published under an MIT license

import yaml
import sys

try:
    from hyddown import HydDown
except:
    import sys
    import os
    hyddown_path = os.path.join(os.path.abspath(os.path.dirname(__file__)),"..","src")
    sys.path.append(os.path.abspath(hyddown_path))

    from hyddown import HydDown


if __name__ == "__main__":
    if len(sys.argv) > 1:
        input_filename = sys.argv[1]
    else:
        input_filename = "input.yml"

    with open(input_filename) as infile:
        input = yaml.load(infile, Loader=yaml.FullLoader)


    hdown=HydDown(input)
    
    hdown.run(disable_pbar=False)
    hdown.plot()


# Thermeshc code adapted verbatim from https://github.com/wjbg/thermesh
#
from __future__ import annotations
import numpy as np
from typing import Callable, Union
import numpy.typing as npt


vector = matrix = npt.NDArray[np.float64]


def solve_ht(domain: Domain, solver: dict[str, float]) -> tuple[vector, matrix]:
    """Solves heat transfer problem.

    Parameters
    ----------
    domain : Domain
        Domain with a mesh, material model and boundary conditions.
    solver : dict
        Solver information, dictionary with the following keys: dt,
        t_end, theta

    Returns
    -------
    t : nd.array(dtype=float, dim=1, len=int(t_end/dt))
        Times.
    T : nd.array(dtype=float, dim=2, shape=(int(t_end/dt), mesh.nn))
        Temperature data.

    """
    timesteps = int(solver["t_end"] / solver["dt"])

    # Data storage
    T = np.zeros((timesteps + 1, domain.mesh.nn))
    t = np.zeros(timesteps + 1)
    T[0] = domain.T
    t[0] = domain.t

    # Solve problem
    for i in range(1, timesteps + 1):
        t[i], T[i] = domain.timestep(solver["dt"], solver["theta"])
    return t, T


class Domain:
    """Class to represent a domain.

    Attributes
    ==========
    mesh : Mesh
        Object with mesh information.
    bc : two-item list of dicts
        The boundary conditions are provided in a two-item list of
        dictionaries. The first dictionary (or zeroth item in the list)
        applies to the start or left side of the domain, while the second
        item applies to the end or right side of the domain. The dictionaries
        can have the keys: "T" OR ( ("h" and "T_inf") AND/OR "q" ), with "T"
        an applied temperature, "h" and "T_inf" the convective heat transfer
        coefficient and far field temperature, respectively, while "q"
        represents a direct flux on the surface which is directed inwards.
    constitutive_model : list[function]
        Functions that takes temperature as an input and provides a
        dictionary with the following keys: k, cp, rho. The list has a
        length equal to the number of subdomains in the mesh. The i-th
        function in the list belongs to the i-th subdomain.
    t : float
        Time.
    T : nd.array(dim=1, len=mesh.nn)
        Temperature at time t.
    q : nd.array(dim=1, len=mesh.nn)
        Heat flux at time t.

    Methods
    =======
    timestep(dt, theta=0.5)
        Apply a timestep dt and update data.

    system_matrices()
        Returns domain stiffness and dampling matrix.

    check_bc()
        Checks if bc's are valid.

    set_T(T) and set_q(q)
        Set temperature and heat flux.

    clear()
        Clears data.

    """

    def __init__(
        self, mesh: Mesh, constitutive_model: list[Callable], bc: dict[str, float] = {}
    ):

        self.mesh = mesh
        self.constitutive_model = constitutive_model
        self.bc = bc

        self.t = 0
        self.T = np.zeros(self.mesh.nn)
        self.q = np.zeros(self.mesh.nn)

    def timestep(self, dt: float, theta: float = 0.5):
        """Apply timestep and update data.

        Parameters
        ----------
        dt : float
            Timestep size.
        theta: float (0 < theta <= 1)
            Timestepping type.

        NOTE: This is far from optimized as each matrix and vector is
              repeatedly constructed.

        NOTE: A more advanced solution strategy should be used when the
              problem becomes nonlinear, e.g. due to a radiative bc or
              due to temperature-dependent material properties.

        NOTE: The internal heat source Q is not implemented yet.

        """
        if self.check_bc():
            K, C = self.system_matrices()
            T_new, q_new = np.zeros(self.mesh.nn), np.zeros(self.mesh.nn)
            H = np.zeros(K.shape)
            j = np.arange(self.mesh.nn)

            for i in [0, -1]:
                if "T" in self.bc[i].keys():
                    T_new[i] = self.bc[i]["T"]
                    j = np.delete(j, i)
                if "q" in self.bc[i].keys():
                    q_new[i] += self.bc[i]["q"]
                if "h" in self.bc[i].keys() and "T_inf" in self.bc[i].keys():
                    H[i, i] = self.bc[i]["h"]
                    q_new[i] += self.bc[i]["h"] * self.bc[i]["T_inf"]

            A = C + dt * theta * (K + H)  # Matrix on LHS
            f = (
                1 - theta
            ) * dt * self.q + dt * theta * q_new  # misses contribution from Q
            b = np.dot(C - dt * (1 - theta) * (K + H), self.T) + f  # RHS
            b -= np.dot(A, T_new)  # take into account T boundary condition
            T_new[j] = np.linalg.solve(A[np.ix_(j, j)], b[j])
            self.T = T_new
            self.q = q_new
            self.t += dt
        return self.t, self.T

    def system_matrices(self) -> tuple[matrix, matrix]:
        """Returns domain stiffness and damping matrix."""
        K = np.zeros((self.mesh.nn, self.mesh.nn))
        C = np.zeros((self.mesh.nn, self.mesh.nn))
        for c, elem, sd in zip(self.mesh._conn, self.mesh.elem, self.mesh.subdomain):
            mat = self.constitutive_model[sd](self.T[c].mean())
            cols = np.tile(c, [len(c), 1])
            rows = np.tile(c[np.newaxis].T, [1, len(c)])
            K[rows, cols] += elem.K(mat)
            C[rows, cols] += elem.C(mat)
        return K, C

    def check_bc(self) -> bool:
        """Check if boundary conditions are valid."""
        for bc in self.bc:
            if "T" in bc.keys():
                if "q" in bc.keys() or "h" in bc.keys():
                    raise KeyError("invalid combination of bc's")
            if "q" in bc.keys():
                if "T" in bc.keys():
                    raise KeyError("invalid combination of bc's")
            if "h" in bc.keys():
                if "T" in bc.keys():
                    raise KeyError("invalid combination of bc's")
                if "T_inf" not in bc.keys():
                    raise KeyError("no temperature provided for convective bc")
        return True

    def set_T(self, T: Union[float, vector]):
        """Set temperature.

        Parameter
        ---------
        T : float OR np.ndarray(dim=1, dtype=float, len=nn)
            Temperature at nodes.

        """
        if type(T) == float:
            self.T = T * np.ones(self.mesh.nn)
        else:
            self.T = T * np.ones(self.mesh.nn)

    def set_q(self, q: Union[float, vector]):
        """Set heat flux.

        Parameter
        ---------
        q : float OR np.ndarray(dim=1, dtype=float, len=nn)
            Heat flux at nodes.

        """
        if type(q) == float:
            self.q = q * np.ones(self.mesh.nn)
        else:
            self.q = q * np.ones(self.mesh.nn)

    def clear(self):
        """Clears time and temperature data."""
        self.t = 0
        self.T = np.zeros(self.mesh.nn)
        self.q = np.zeros(self.mesh.nn)


def isothermal_model(k: float, rho: float, cp: float) -> Callable:
    """Returns a function that represents an isothermal material model.

    Parameter
    ---------
    k : float
        Thermal conductivity.
    rho : float
        Density
    cp : float
        Specific heat.

    Returns
    -------
    model : Callable
        Function that returns a dictionary with the provided
        constitutive properties.

    """

    def model(T: float) -> dict[str, float]:
        """A isothermal constitutive model.

        Parameter
        ---------
        T : float
            Temperature

        Returns
        -------
        mat : dict[str, float]
            Dictionary with the following keys: k, cp, rho.

        """
        return {"k": k, "rho": rho, "cp": cp}

    return model


def piecewise_linear_model(k: Matrix, rho: Matrix, cp: Matrix) -> Callable:
    """Returns a function that represents an isothermal material model.

    Parameter
    ---------
    k : np.ndarray(dim=2, dtype=float)
        Temperature vs. thermal conductivity.
    rho : np.ndarray(dim=2, dtype=float)
        Temperature vs. density
    cp : np.ndarray(dim=2, dtype=float)
        Temperature vs. specific heat.

    Returns
    -------
    model : Callable
        Function that returns a dictionary with the provided
        constitutive properties.

    """

    def model(T: float) -> dict[str, float]:
        """An piece-wise linear constitutive model.

        Parameter
        ---------
        T : float
            Temperature

        Returns
        -------
        mat : dict[str, float]
            Dictionary with the following keys: k, rho, cp.

        """
        k_ = np.interp(T, k[:, 0], k[:, 1])  # conductivity at T
        rho_ = np.interp(T, rho[:, 0], rho[:, 1])
        cp_ = np.interp(T, cp[:, 0], cp[:, 1])
        return {"k": k_, "rho": rho_, "cp": cp_}

    return model


class Mesh:
    """Class to represent a mesh.

    Attributes
    ==========
    nodes : nd.array()
        Node locations.
    elem : list[Element]
        List of elements.
    nn : int
        Number of nodes.
    nel : int
        Number of elements.
    subdomain : list[int] (defaults to a list with zeros)
        Number that indicates the subdomain in the mesh. Correlates
        with the constitutive model that will be used for the
        analysis.

    """

    def __init__(self, z: vector, element: LinearElement):
        """Initializes Mesh instance.

        Parameters
        ----------
        z : np.ndarray(dim=1, dtype=float)
            Node locations.
        element : Element
            Element type.

        """
        if (len(z) - 1) % (element.order) == 0:
            self.nn = len(z)
            self.nodes = z
            self._conn = np.array(
                [
                    np.arange(i, i + 1 + element.order)
                    for i in np.arange(0, self.nn - 1, element.order)
                ]
            )
            self.elem = [element(self.nodes[c]) for c in self._conn]
            self.nel = len(self.elem)
            self.subdomain = [0] * self.nel
        else:
            raise ValueError("node and element number mismatch")

    def __str__(self) -> str:
        s = "Mesh information\n"
        s += "----------------\n"
        s += f"Nodes: {self.nn:<3}\n"
        s += f"Elements: {self.nel:<3}\n"
        s += f"Element type: {type(self.elem[0])}\n"
        return s


class Element:
    """Class to represent an element.

    Attributes
    ==========
    order : int
        Order of the element.
    dim : int
        Dimension of the element.

    """

    dim = 1
    order = 0

    def __init__(self, nodes: vector):
        """Initializes Element instance.

        Parameters
        ----------
        nodes : np.ndarray(dim=1, dtype=float)
            Node locations.

        """
        self.nodes = nodes

    def length(self) -> float:
        return self.nodes[-1] - self.nodes[0]


class LinearElement(Element):
    """Class to represent a linear element (order = 1).

    Methods
    =======
    K()
        Returns element stiffness matrix.
    C()
        Returns element damping matrix.

    """

    order = 1

    def K(self, mat: dict[str, float]) -> matrix:
        """Returns element stiffness matrix."""
        K = (mat["k"] / self.length()) * np.array([[1, -1], [-1, 1]])
        return K

    def C(self, mat: dict[str, float]):
        """Returns element damping matrix."""
        C = (self.length() * mat["rho"] * mat["cp"] / 6) * np.array([[2, 1], [1, 2]])
        return C


class QuadraticElement(Element):
    """Class to represent a linear element (order = 2).

    Methods
    =======
    K()
        Returns element stiffness matrix.
    C()
        Returns element damping matrix.

    """

    order = 2

    def K(self, mat: dict[str, float]) -> matrix:
        """Returns element stiffness matrix."""
        K = (mat["k"] / self.length() / 3) * np.array(
            [[7, -8, 1], [-8, 16, -8], [1, -8, 7]]
        )
        return K

    def C(self, mat: dict[str, float]) -> matrix:
        """Returns element damping matrix."""
        C = (self.length() * mat["rho"] * mat["cp"] / 30) * np.array(
            [[4, 2, -1], [2, 16, 2], [-1, 2, 4]]
        )
        return C



# HydDown hydrogen/other gas depressurisation
# Copyright (c) 2021 Anders Andreasen
# Published under an MIT license

from hyddown import transport as tp
from hyddown import fire
from hyddown import validator
from CoolProp.CoolProp import PropsSI
import CoolProp.CoolProp as CP
import pytest


def get_example_input(fname):
    import os
    import yaml

    fname = os.path.join(os.path.abspath(os.path.dirname(__file__)), "examples", fname)

    with open(fname) as infile:
        input = yaml.load(infile, Loader=yaml.FullLoader)

    return input


def test_orifice():
    P1 = 10.0e5
    P2 = 5.5e5
    D = PropsSI("D", "P", P1, "T", 298.15, "HEOS::N2")
    cpcv = PropsSI("CP0MOLAR", "T", 298.15, "P", P1, "HEOS::N2") / PropsSI(
        "CVMOLAR", "T", 298.15, "P", P1, "HEOS::N2"
    )
    assert tp.gas_release_rate(
        P1, P2, D, cpcv, 0.85, 0.01**2 / 4 * 3.1415
    ) == pytest.approx(9.2 / 60, rel=0.001)


def test_orifice1():
    P1 = 10.0e5
    P2 = 6.5e5
    D = PropsSI("D", "P", P1, "T", 298.15, "HEOS::N2")
    cpcv = PropsSI("CP0MOLAR", "T", 298.15, "P", P1, "HEOS::N2") / PropsSI(
        "CVMOLAR", "T", 298.15, "P", P1, "HEOS::N2"
    )
    assert tp.gas_release_rate(
        P1, P2, D, cpcv, 0.85, 0.01**2 / 4 * 3.1415
    ) == pytest.approx(9.2 / 60, rel=0.2)


def test_controlvalve():
    P1 = 10.0e5
    P2 = 5.5e5
    T1 = 20.0 + 273.15
    MW = PropsSI("M", "P", P1, "T", T1, "HEOS::N2")
    Z1 = PropsSI("Z", "P", P1, "T", T1, "HEOS::N2")
    gamma = PropsSI("CP0MOLAR", "T", T1, "P", P1, "HEOS::N2") / PropsSI(
        "CVMOLAR", "T", T1, "P", P1, "HEOS::N2"
    )
    assert tp.control_valve(P1, P2, T1, Z1, MW, gamma, 500) == pytest.approx(
        21.92, rel=0.05
    )


def test_cv_vs_time():
    assert tp.cv_vs_time(1, 0.5, time_constant=1, characteristic="linear") == 0.5
    assert tp.cv_vs_time(1, 0.5, time_constant=1, characteristic="eq") == pytest.approx(
        0.14, abs=0.002
    )
    assert tp.cv_vs_time(
        1, 0.5, time_constant=1, characteristic="fast"
    ) == pytest.approx(0.707, abs=0.002)
    assert tp.cv_vs_time(1, 0.5, time_constant=0) == 1.0


def test_psv3():
    Pback = 1e5
    Pset = 18.2e5
    blowdown = 0.1
    P1 = 0.99 * Pset * (1 - blowdown)
    T1 = 100.0 + 273.15
    Z = PropsSI("Z", "P", P1, "T", T1, "HEOS::N2")
    MW = PropsSI("M", "HEOS::N2")
    gamma = PropsSI("CP0MOLAR", "T", T1, "P", P1, "HEOS::N2") / PropsSI(
        "CVMOLAR", "T", T1, "P", P1, "HEOS::N2"
    )
    CD = 0.975
    area = 71e-6
    assert tp.relief_valve(P1, Pback, Pset, blowdown, gamma, CD, T1, Z, MW, area) == 0
    psv_state = "closed"
    assert (
        tp.relief_valve(P1 * 1.01, Pback, Pset, blowdown, gamma, CD, T1, Z, MW, area)
        == 0
    )


def test_psv2():
    P1 = 21.0e5
    Pback = 1e5
    Pset = 20.99e5
    blowdown = 0.1
    T1 = 100.0 + 273.15
    Z = PropsSI("Z", "P", P1, "T", T1, "HEOS::N2")
    MW = PropsSI("M", "HEOS::N2")
    gamma = PropsSI("CP0MOLAR", "T", T1, "P", P1, "HEOS::N2") / PropsSI(
        "CVMOLAR", "T", T1, "P", P1, "HEOS::N2"
    )
    CD = 0.975
    area = 71e-6
    assert tp.relief_valve(
        P1, Pback, Pset, blowdown, gamma, CD, T1, Z, MW, area
    ) == pytest.approx(1046 / 3600, rel=0.03)
    psv_state = "open"
    assert tp.relief_valve(
        Pset * 0.99, Pback, Pset, blowdown, gamma, CD, T1, Z, MW, area
    ) == pytest.approx(1046 / 3600, rel=0.03)


def test_psv():
    P1 = 100e5
    Pback = 1e5
    Pset = 99.2e5
    blowdown = 0.1
    T1 = 25.0 + 273.15
    Z = PropsSI("Z", "P", P1, "T", T1, "HEOS::N2")
    MW = PropsSI("M", "HEOS::N2")
    gamma = PropsSI("CP0MOLAR", "T", T1, "P", P1, "HEOS::N2") / PropsSI(
        "CVMOLAR", "T", T1, "P", P1, "HEOS::N2"
    )
    CD = 0.975
    area = 71e-6
    assert tp.relief_valve(
        P1, Pback, Pset, blowdown, gamma, CD, T1, Z, MW, area
    ) == pytest.approx(1.57, rel=0.02)


def test_api_psv_relief():
    assert tp.api_psv_release_rate(
        121.9e5, 71e5, 1.39, 0.975, 298.15, 1.01, 2 / 1e3, 71e-6
    ) == pytest.approx(1846 / 3600, rel=0.01)
    assert tp.api_psv_release_rate(
        121.9e5, 1e5, 1.39, 0.975, 298.15, 1.01, 2 / 1e3, 71e-6
    ) == pytest.approx(1860 / 3600, rel=0.01)


def test_hinside():
    fluid = CP.AbstractState("HEOS", "air")
    Tboundary = (311 + 505.4) / 2
    fluid.update(CP.PT_INPUTS, 1e5, Tboundary)
    h = tp.h_inside(0.305, 311, 505.4, fluid)
    assert h == pytest.approx(7, abs=0.1)


def test_hinside_mixed():
    mdot = 1e-10
    D = 0.010
    fluid = CP.AbstractState("HEOS", "air")
    Tboundary = (311 + 505.4) / 2
    fluid.update(CP.PT_INPUTS, 1e5, Tboundary)
    h = tp.h_inside_mixed(0.305, 311, 505.4, fluid, mdot, D)
    assert h == pytest.approx(7, abs=0.1)


def test_hinner():
    h = tp.h_inner(0.305, 311, 505.4, 1e5, "HEOS::air")
    assert h == pytest.approx(7, abs=0.1)


def test_hinner_mixed():
    mdot = 1e-10
    D = 0.010
    h = tp.h_inner_mixed(0.305, 311, 505.4, 1e5, "HEOS::air", mdot, D)
    assert h == pytest.approx(7, abs=0.1)


def test_NNu():
    NGr = tp.Gr(0.305, 311, 505.4, 1e5, "HEOS::air")
    NPr = tp.Pr((311 + 505.4) / 2, 1e5, "HEOS::air")
    NRa = NGr * NPr
    NNu = tp.Nu(NRa, NPr)
    print(NNu)
    assert NNu == pytest.approx(62.2, abs=1.0)


def test_NRa():
    NGr = tp.Gr(0.305, 311, 505.4, 1e5, "HEOS::air")
    NPr = tp.Pr((311 + 505.4) / 2, 1e5, "HEOS::air")
    assert (NGr * NPr) == pytest.approx(1.3e8, abs=0.1e8)


def test_NGr():
    assert tp.Gr(0.305, 311, 505.4, 1e5, "HEOS::air") == pytest.approx(1.8e8, abs=0.1e8)


def test_NPr():
    assert tp.Pr((311 + 505.4) / 2, 1e5, "HEOS::air") == pytest.approx(0.7, rel=0.01)


def test_stefan_boltzmann():
    alpha = 1
    e_flame = 1
    e_surface = 0
    h = 100
    Tflame = 635 + 273.15
    Tradiative = 635 + 273.15
    assert (
        fire.stefan_boltzmann(
            alpha, e_flame, e_surface, h, Tflame, Tradiative, 20 + 273.15
        )
    ) == pytest.approx(1e5, abs=100)


def test_pool_fire_api521():
    assert fire.pool_fire_api521(273 + 50) == pytest.approx(45.5e3, abs=100)


def test_jet_fire_api521():
    assert fire.jet_fire_api521(273 + 50) == pytest.approx(83.5e3, abs=500)


def test_jet_fire_scandpower():
    assert fire.jet_fire_scandpower(273 + 20) == pytest.approx(94.5e3, abs=1000)


def test_pool_fire_scandpower():
    assert fire.pool_fire_scandpower(273 + 20) == pytest.approx(88.5e3, abs=500)


def test_sb():
    assert fire.sb_fire(273 + 50, "api_jet") == pytest.approx(83.5e3, abs=500)
    assert fire.sb_fire(273 + 50, "api_pool") == pytest.approx(45.5e3, abs=100)
    assert fire.sb_fire(273 + 20, "scandpower_pool") == pytest.approx(88.5e3, abs=500)
    assert fire.sb_fire(273 + 20, "scandpower_jet") == pytest.approx(94.5e3, abs=1000)
    try:
        Q = fire.sb_fire(273 + 20, "scand_jet") == pytest.approx(94.5e3, abs=1000)
    except ValueError:
        pass


def test_validator():
    import os
    import yaml

    dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), "examples")

    for fname in os.listdir(dir):
        with open(os.path.join(dir, fname)) as infile:
            input = yaml.load(infile, Loader=yaml.FullLoader)
        assert validator.validate_mandatory_ruleset(input) == True


def test_validator2():
    import os
    import yaml

    dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), "examples")

    for fname in os.listdir(dir):
        with open(os.path.join(dir, fname)) as infile:
            input = yaml.load(infile, Loader=yaml.FullLoader)
        assert validator.heat_transfer_validation(input) != False
        assert validator.valve_validation(input) != False


def test_sim_orifice_full():
    from hyddown import HydDown

    input = get_example_input("input.yml")
    hdown = HydDown(input)
    hdown.run()
    hdown.plot(verbose=False)
    hdown.generate_report()
    assert hdown.report["final_mass"] == pytest.approx(0.14, rel=0.02)
    assert hdown.report["min_wall_temp"] == pytest.approx(284.9, rel=0.01)
    assert hdown.report["min_fluid_temp"] == pytest.approx(193.7, rel=0.01)
    assert hdown.report["time_min_fluid_temp"] == pytest.approx(37.0, rel=0.01)
    assert hdown.report["max_mass_rate"] == pytest.approx(0.88, rel=0.01)


def test_sim_controlvalve():
    from hyddown import HydDown

    input = get_example_input("controlvalve.yml")
    hdown = HydDown(input)
    hdown.run()


def test_sim_psv():
    from hyddown import HydDown

    input = get_example_input("psv.yml")
    hdown = HydDown(input)
    hdown.run()


def test_isenthalpic():
    from hyddown import HydDown

    input = get_example_input("isenthalpic.yml")
    hdown = HydDown(input)
    hdown.run()


def test_sim_cv_filling():
    from hyddown import HydDown

    input = get_example_input("cv_filling.yml")
    hdown = HydDown(input)
    hdown.run()


def test_mdot_filling():
    from hyddown import HydDown

    input = get_example_input("mdot_filling.yml")
    hdown = HydDown(input)
    hdown.run()


def test_sim_filling():
    from hyddown import HydDown

    input = get_example_input("filling.yml")
    hdown = HydDown(input)
    hdown.run()


def test_multicomponent():
    from hyddown import HydDown

    input = get_example_input("ng.yml")
    hdown = HydDown(input)
    hdown.run()


def test_sim_stefan_boltzmann():
    from hyddown import HydDown

    input = get_example_input("psv_sb.yml")
    hdown = HydDown(input)
    hdown.run()


def test_dataframe():
    from hyddown import HydDown

    input = get_example_input("controlvalve.yml")
    hdown = HydDown(input)
    hdown.run()
    df = hdown.get_dataframe()


def test_thermesh():
    from hyddown import thermesh as tm

    import numpy as np
    import matplotlib.pyplot as plt
    from scipy.special import erfc

    # Domain information
    L = 0.01  # Make sure the domain is large enough to be a semi-infinite solid
    k, rho, cp = 0.72, 1560.0, 1450.0
    h = 20.0  # Heat transfer coefficient
    T_0, T_inf = 0.0, 400.0  # Initial and far field temperature

    #
    # Analytical solution
    # -------------------

    def analytical_solution(x, t, rho, cp, k, q, T_inf):
        """Returns analytical solution when initial temperature equals zero.

        Arguments
        ---------
        x : nd.array(dtype=float, dim=1)
            Spatial coordinates.
        t : float
            Time.
        rho : float
            Density.
        cp : float
            Specific heat.
        k : float
            Thermal conductivity
        h : float
            Heat transfer coefficient.
        T_inf : float
            Far field temperature.

        Returns
        -------
        dT : nd.array(dtype=float, dim=1, len=len(x))
            Temperature change at x.

        """
        alpha = k / rho / cp
        dT = T_inf * (
            erfc(x / (2 * np.sqrt(alpha * t)))
            - np.exp(h * x / k + h**2 * alpha * t / k**2)
            * erfc(x / (2 * np.sqrt(alpha * t)) + h * np.sqrt(alpha * t) / k)
        )
        return dT

    # Plot analytical solution at the following times
    t_inc = np.array([2, 10, 25])

    # Domain coordinates and thermal diffusivity
    z = np.linspace(0, L, 50)

    fig = plt.figure(figsize=(2.4, 2.0))
    ax = fig.add_subplot(1, 1, 1)
    ax.set_title("Analytical solution")
    ax.set_xlabel("z/L")
    ax.set_ylabel(r"$\Delta$T ($^{\circ}$C)")
    ax.set_xlim([0, 1])
    for t in t_inc:
        plt.plot(
            z / L,
            analytical_solution(z, t, rho, cp, k, h, T_inf),
            "k",
            linewidth=0.5,
        )
    plt.text(0.04, 1, "2")
    plt.text(0.12, 5, "10")
    plt.text(0.40, 8, "25")
    plt.tight_layout()
    # plt.savefig("../fig/conv_analytical_sol.png", dpi=600)
    # plt.show(block=False)

    #
    # Numerical solution
    # ------------------

    # Let's make a mesh using linear elements (LinearElement). The
    # alternative is to use second-order elements (QuadraticElement).
    nn = 11  # number of nodes
    z = np.linspace(0, L, nn)
    mesh = tm.Mesh(z, tm.LinearElement)  # Or `QuadraticElement` to
    # use quadratic shape functions

    # The boundary conditions are provided in a two-item list of
    # dictionaries. The first dictionary (or zeroth item in the list)
    # applies to the start or left side of the domain, while the second
    # item applies to the end or right side of the domain. The
    # dictionaries can have the following keys:
    #
    #   "T" OR ( ("h" and "T_inf") AND/OR "q" ),
    #
    # with "T" an applied temperature, "h" and "T_inf" the convective heat
    # transfer coefficient and far field temperature, respectively, while
    # "q" represents a direct flux on the surface.
    bc = [
        {"h": h, "T_inf": T_inf},  # convective bc on the left
        {"T": T_0},
    ]  # T on the right

    # Material model (CPEEK is a function that takes T as input)
    cpeek = tm.isothermal_model(k, rho, cp)

    # Define and solve problem
    domain = tm.Domain(mesh, [cpeek], bc)
    domain.set_T(T_0 * np.ones(nn))

    # Solver details
    theta = 0.5
    dt = 0.5
    solver = {"dt": dt, "t_end": 25.0, "theta": theta}
    t, T = tm.solve_ht(domain, solver)

    # Plot numerical solution
    fig = plt.figure(figsize=(2.4, 2.0))
    ax = fig.add_subplot(1, 1, 1)
    ax.set_title(
        "FE solution (dz/L="
        + str((z[1] - z[0]) / L)
        + ", dt="
        + str(dt)
        + r" s., $\Theta$="
        + str(theta)
        + ")"
    )
    ax.set_xlabel("z/L")
    ax.set_ylabel(r"$\Delta$T ($^{\circ}$C)")
    ax.set_xlim([0, 1])
    t_inc = np.array([2, 10, 25]) / dt
    for i in t_inc.astype(int):
        plt.plot(z / L, T[i, :], "k", linewidth=0.5)
    plt.text(0.04, 1, "2")
    plt.text(0.12, 5, "10")
    plt.text(0.40, 8, "25")
    plt.tight_layout()
    # plt.savefig("../fig/conv_FE_t0.5_dt0.1s.png", dpi=600)
    # plt.show(block=False)
    analytical = analytical_solution(z, 25, rho, cp, k, h, T_inf)
    print(T[-1, :])
    print(analytical)
    print(sum((T[-1, :] - analytical) ** 2))

    assert T[-1, 0] == pytest.approx(analytical[0], abs=0.3)
    assert sum((T[-1, :] - analytical) ** 2) < 1


if __name__ == "__main__":
    test_thermesh()



# HydDown hydrogen/other gas depressurisation
# Copyright (c) 2021 Anders Andreasen
# Published under an MIT license


def stefan_boltzmann(alpha, e_flame, e_surface, h, Tflame, Tradiative, Tvessel):
    """
    Heat flux from flame to vessel surface from combined radiative
    and convective heat transfer. Returns the heat flux in W/m2
    See also API 521 and Scandpower guidelines.
    """
    sigma = 5.67e-8  # Stefan-Botlzmann constant W/m^2 K^4
    return (
        alpha * e_flame * sigma * Tradiative ** 4
        + h * (Tflame - Tvessel)
        - e_surface * sigma * Tvessel ** 4
    )


def pool_fire_api521(Tvessel):
    """
    Incident heat flux of 60 kW/m2
    """
    alpha = 0.75
    e_flame = 0.75
    e_surface = 0.75
    h = 20
    Tflame = 600 + 273.15
    Tradiative = 750 + 273.15
    return stefan_boltzmann(alpha, e_flame, e_surface, h, Tflame, Tradiative, Tvessel)


def pool_fire_scandpower(Tvessel):
    """
    Incident heat flux of 100 kW/m2
    """
    alpha = 0.85
    e_flame = 1
    e_surface = 0.85
    h = 30
    Tflame = 804 + 273.15
    Tradiative = 804 + 273.15
    return stefan_boltzmann(alpha, e_flame, e_surface, h, Tflame, Tradiative, Tvessel)


def jet_fire_api521(Tvessel):
    """
    Incident heat flux of 100 kW/m2
    """
    alpha = 0.75
    e_flame = 0.33
    e_surface = 0.75
    h = 40
    Tflame = 900 + 273.15
    Tradiative = 1100 + 273.15
    return stefan_boltzmann(alpha, e_flame, e_surface, h, Tflame, Tradiative, Tvessel)


def jet_fire_scandpower(Tvessel):
    """
    Incident heat flux of 100 kW/m2
    """
    alpha = 0.85
    e_flame = 1
    e_surface = 0.85
    h = 100
    Tflame = 635 + 273.15
    Tradiative = 635 + 273.15
    return stefan_boltzmann(alpha, e_flame, e_surface, h, Tflame, Tradiative, Tvessel)


def sb_fire(T_vessel, fire_type):
    if fire_type == "api_jet":
        Q=jet_fire_api521(T_vessel)
    elif fire_type == "api_pool":
        Q=pool_fire_api521(T_vessel)
    elif fire_type == "scandpower_pool":
        Q=pool_fire_scandpower(T_vessel)
    elif fire_type == "scandpower_jet":
        Q=jet_fire_scandpower(T_vessel)
    else:
        raise ValueError("Unknown Stefan-Bolzmann fire heat load")
    return Q




# HydDown hydrogen/other gas depressurisation
# Copyright (c) 2021 Anders Andreasen
# Published under an MIT license

import math
import numpy as np
import pandas as pd
from tqdm import tqdm
from scipy.optimize import minimize
from CoolProp.CoolProp import PropsSI
import CoolProp.CoolProp as CP
from hyddown import transport as tp
from hyddown import validator
from hyddown import fire
from hyddown import thermesh as tm


class HydDown:
    """
    Main class to to hold problem definition, running problem, storing results etc.
    """

    def __init__(self, input):
        """
        Parameters
        ----------
        input : dict
            Dict holding problem definition
        """
        self.input = input
        self.verbose = 0
        self.isrun = False
        self.validate_input()
        self.read_input()
        self.initialize()

    def validate_input(self):
        """
        Validating the provided problem definition dict

        Raises
        ------
        ValueError
            If missing input is detected.
        """
        valid = validator.validation(self.input)
        if valid is False:
            raise ValueError("Error in input file")

    def read_input(self):
        """
        Reading in input/ problem definition dict and assigning to classs
        attributes.
        """
        self.length = self.input["vessel"]["length"]
        self.diameter = self.input["vessel"]["diameter"]

        self.p0 = self.input["initial"]["pressure"]
        self.T0 = self.input["initial"]["temperature"]

        self.species = "HEOS::" + self.input["initial"]["fluid"]

        # Detects if a multi component fluid is specified using & for separation of components
        if "&" in self.input["initial"]["fluid"]:
            comp_frac_pair = [
                str.replace("[", " ").replace("]", "").split(" ")
                for str in self.input["initial"]["fluid"].split("&")
            ]
            comp = [pair[0] for pair in comp_frac_pair]
            compSRK = [pair[0] + "-SRK" for pair in comp_frac_pair]
            molefracs = np.asarray([float(pair[1]) for pair in comp_frac_pair])
            molefracs = molefracs / sum(molefracs)
            self.molefracs = molefracs
            sep = "&"
            self.comp = sep.join(comp)
            self.compSRK = sep.join(compSRK)
        # Normally single component fluid is specified
        else:
            self.comp = self.input["initial"]["fluid"]
            self.molefracs = [1.0]
            self.compSRK = self.input["initial"]["fluid"]

        self.tstep = self.input["calculation"]["time_step"]
        self.time_tot = self.input["calculation"]["end_time"]
        self.method = self.input["calculation"]["type"]

        # Reading valve specific data
        if (
            self.input["valve"]["type"] == "orifice"
            or self.input["valve"]["type"] == "psv"
        ):
            self.p_back = self.input["valve"]["back_pressure"]
            self.D_orifice = self.input["valve"]["diameter"]
            self.CD = self.input["valve"]["discharge_coef"]
            if self.input["valve"]["type"] == "psv":
                self.Pset = self.input["valve"]["set_pressure"]
                self.blowdown = self.input["valve"]["blowdown"]
                self.psv_state = "closed"
        elif self.input["valve"]["type"] == "controlvalve":
            self.p_back = self.input["valve"]["back_pressure"]
            self.Cv = self.input["valve"]["Cv"]
            if "xT" in self.input["valve"]:
                self.xT = self.input["valve"]["xT"]
            if "Fp" in self.input["valve"]:
                self.Fp = self.input["valve"]["Fp"]
            if (
                "characteristic" in self.input["valve"]
                and "time_constant" in self.input["valve"]
            ):
                self.valve_characteristic = self.input["valve"]["characteristic"]
                self.valve_time_constant = self.input["valve"]["time_constant"]
            else:
                self.valve_characteristic = "linear"
                self.valve_time_constant = 0

        elif (
            self.input["valve"]["type"]
            == "mdot"
            # and self.input["valve"]["flow"] == "filling"
        ):
            self.p_back = self.input["valve"]["back_pressure"]

        # valve type
        # - constant_mass
        # - functional mass flow
        self.thickness = 0

        # Reading heat transfer related data/information
        if "heat_transfer" in self.input:
            self.heat_method = self.input["heat_transfer"]["type"]
            if self.heat_method == "specified_h" or self.heat_method == "specified_U":
                self.Tamb = self.input["heat_transfer"]["temp_ambient"]
            if self.heat_method == "specified_U":
                self.Ufix = self.input["heat_transfer"]["U_fix"]
            if self.heat_method == "specified_Q":
                self.Qfix = self.input["heat_transfer"]["Q_fix"]
            if self.heat_method == "specified_h":
                self.vessel_cp = self.input["vessel"]["heat_capacity"]
                self.vessel_density = self.input["vessel"]["density"]
                self.vessel_orientation = self.input["vessel"]["orientation"]
                self.thickness = self.input["vessel"]["thickness"]
                self.h_out = self.input["heat_transfer"]["h_outer"]
                self.h_in = self.input["heat_transfer"]["h_inner"]
                if self.input["valve"]["flow"] == "filling":
                    if "D_throat" in self.input["heat_transfer"]:
                        self.D_throat = self.input["heat_transfer"]["D_throat"]
                    else:
                        self.D_throat = self.input["vessel"]["diameter"]
            if self.heat_method == "s-b":
                self.fire_type = self.input["heat_transfer"]["fire"]
                self.h_in = "calc"
                self.vessel_cp = self.input["vessel"]["heat_capacity"]
                self.vessel_density = self.input["vessel"]["density"]
                self.vessel_orientation = self.input["vessel"]["orientation"]
                self.thickness = self.input["vessel"]["thickness"]
                if self.input["valve"]["flow"] == "filling":
                    raise ValueError("Filling and Fire heat load not implemented")

    def initialize(self):
        """
        Preparing for running problem by creating the fluid objects required
        instantiating arrays for storing time-dependent results, setting additional
        required class attributes.
        """
        self.vol = self.diameter**2 / 4 * math.pi * self.length  # m3
        self.vol_tot = (
            (self.diameter + 2 * self.thickness) ** 2
            / 4
            * math.pi
            * (self.length + 2 * self.thickness)
        )  # m3
        self.vol_solid = self.vol_tot - self.vol
        self.surf_area_outer = (
            self.diameter + 2 * self.thickness
        ) ** 2 / 4 * math.pi * 2 + (self.diameter + 2 * self.thickness) * math.pi * (
            self.length + 2 * self.thickness
        )
        self.surf_area_inner = (self.diameter) ** 2 / 4 * math.pi * 2 + (
            self.diameter
        ) * math.pi * self.length

        self.fluid = CP.AbstractState("HEOS", self.comp)
        self.fluid.specify_phase(CP.iphase_gas)
        self.fluid.set_mole_fractions(self.molefracs)
        self.fluid.update(CP.PT_INPUTS, self.p0, self.T0)

        self.transport_fluid = CP.AbstractState("HEOS", self.compSRK)
        self.transport_fluid.specify_phase(CP.iphase_gas)
        self.transport_fluid.set_mole_fractions(self.molefracs)

        self.vent_fluid = CP.AbstractState("HEOS", self.comp)
        self.vent_fluid.specify_phase(CP.iphase_gas)
        self.vent_fluid.set_mole_fractions(self.molefracs)
        self.vent_fluid.update(CP.PT_INPUTS, self.p0, self.T0)

        self.res_fluid = CP.AbstractState("HEOS", self.comp)
        self.res_fluid.set_mole_fractions(self.molefracs)
        if self.input["valve"]["flow"] == "filling":
            self.res_fluid.update(CP.PT_INPUTS, self.p_back, self.T0)

        # data storage
        data_len = int(self.time_tot / self.tstep)
        self.rho = np.zeros(data_len)
        self.T_fluid = np.zeros(data_len)
        self.T_vent = np.zeros(data_len)
        self.T_vessel = np.zeros(data_len)
        self.T_inner_wall = np.zeros(data_len)
        self.T_outer_wall = np.zeros(data_len)
        self.T_bonded_wall = np.zeros(data_len)
        self.Q_outer = np.zeros(data_len)
        self.Q_inner = np.zeros(data_len)
        self.h_inside = np.zeros(data_len)
        self.T_vent = np.zeros(data_len)
        self.H_mass = np.zeros(data_len)
        self.S_mass = np.zeros(data_len)
        self.U_mass = np.zeros(data_len)
        self.U_tot = np.zeros(data_len)
        self.U_res = np.zeros(data_len)
        self.P = np.zeros(data_len)
        self.mass_fluid = np.zeros(data_len)
        self.mass_rate = np.zeros(data_len)
        self.time_array = np.zeros(data_len)
        self.temp_profile = []
        self.rho0 = (
            self.fluid.rhomass()
        )  # PropsSI("D", "T", self.T0, "P", self.p0, self.species)
        self.m0 = self.rho0 * self.vol
        self.MW = self.fluid.molar_mass()  # PropsSI("M", self.species)

    def PHres(self, T, P, H):
        """
        Residual enthalpy function to be minimised during a PH-problem

        Parameters
        ----------
        H : float
            Enthalpy at initial/final conditions
        P : float
            Pressure at final conditions.
        T : float
            Updated estimate for the final temperature at P,H
        """
        self.vent_fluid.update(CP.PT_INPUTS, P, T)
        return ((H - self.vent_fluid.hmass()) / H) ** 2

    def PHproblem(self, H, P, Tguess):
        """
        Defining a constant pressure, constant enthalpy problem i.e. typical adiabatic
        problem like e.g. valve flow for the vented flow (during discharge).
        For multicomponent mixture the final temperature is changed/optimised until the residual
        enthalpy is near zero in an optimisation step. For single component fluid the coolprop
        built in methods are used for speed.

        Parameters
        ----------
        H : float
            Enthalpy at initial/final conditions
        P : float
            Pressure at final conditions.
        Tguess : float
            Initial guess for the final temperature at P,H
        """

        # Multicomponent case
        if "&" in self.species:
            x0 = Tguess
            res = minimize(
                self.PHres,
                x0,
                args=(P, H),
                method="Nelder-Mead",
                options={"xatol": 0.1, "fatol": 0.001},
            )
            T1 = res.x[0]
        # single component fluid case
        else:
            T1 = PropsSI("T", "P", P, "H", H, self.species)
        return T1

    def UDres(self, x, U, rho):
        """
        Residual U-rho to be minimised during a U-rho/UV-problem

        Parameters
        ----------
        U : float
            Internal energy at final conditions
        rho : float
            Density at final conditions
        """
        self.fluid.update(CP.PT_INPUTS, x[0], x[1])
        return ((U - self.fluid.umass()) / U) ** 2 + (
            (rho - self.fluid.rhomass()) / rho
        ) ** 2

    def UDproblem(self, U, rho, Pguess, Tguess):
        """
        Defining a constant UV problem i.e. constant internal energy and density/volume
        problem relevant for the 1. law of thermodynamics.
        For multicomponent mixture the final temperature/pressure is changed/optimised until the
        residual U/rho is near zero. For single component fluid the coolprop
        built in methods are used for speed.

        Parameters
        ----------
        U : float
            Internal energy at final conditions
        rho : float
            Density at final conditions.
        Pguess : float
            Initial guess for the final pressure at U, rho
        Tguess : float
            Initial guess for the final temperature at U, rho
        """
        if "&" in self.species:
            x0 = [Pguess, Tguess]
            res = minimize(
                self.UDres,
                x0,
                args=(U, rho),
                method="Nelder-Mead",
                options={"xatol": 0.1, "fatol": 0.001},
            )
            P1 = res.x[0]
            T1 = res.x[1]
            Ures = U - self.fluid.umass()
        else:
            P1 = PropsSI("P", "D", rho, "U", U, self.species)
            T1 = PropsSI("T", "D", rho, "U", U, self.species)
            Ures = 0
        return P1, T1, Ures

    def run(self, disable_pbar=True):
        """
        Routine for running the actual problem defined i.e. integrating the mass and energy balances
        """
        # Inititialise / setting initial values for t=0
        if self.isrun is True:
            self.initialize()
        input = self.input
        self.rho[0] = self.rho0
        self.T_fluid[0] = self.T0
        self.T_vessel[0] = self.T0
        self.T_inner_wall[0] = self.T0
        self.T_outer_wall[0] = self.T0
        self.T_bonded_wall[0] = self.T0
        if self.input["valve"]["flow"] == "discharge":
            self.T_vent[0] = self.T0
        self.H_mass[0] = self.fluid.hmass()
        self.S_mass[0] = self.fluid.smass()
        self.U_mass[0] = self.fluid.umass()
        self.U_tot[0] = self.fluid.umass() * self.m0
        self.P[0] = self.p0
        self.mass_fluid[0] = self.m0
        cpcv = self.fluid.cp0molar() / (self.fluid.cp0molar() - 8.314)
        massflow_stop_switch = 0

        # Calculating initial mass rate for t=0 depending on mass flow device
        # and filling/discharge mode
        if input["valve"]["type"] == "orifice":
            if input["valve"]["flow"] == "filling":
                k = self.res_fluid.cp0molar() / (self.res_fluid.cp0molar() - 8.314)
                self.mass_rate[0] = -tp.gas_release_rate(
                    self.p_back,
                    self.p0,
                    self.res_fluid.rhomass(),
                    k,
                    self.CD,
                    self.D_orifice**2 / 4 * math.pi,
                )
            else:
                self.mass_rate[0] = tp.gas_release_rate(
                    self.p0,
                    self.p_back,
                    self.rho0,
                    cpcv,
                    self.CD,
                    self.D_orifice**2 / 4 * math.pi,
                )
        elif input["valve"]["type"] == "mdot":
            if "mdot" in input["valve"].keys() and "time" in input["valve"].keys():
                mdot = np.asarray(input["valve"]["mdot"])
                time = np.asarray(input["valve"]["time"])
                max_i = int(time[-1] / self.tstep)
                interp_time = np.linspace(
                    0,
                    self.tstep * len(self.time_array),
                    len(self.time_array),
                    endpoint=False,
                )[:max_i]
                self.mass_rate[:max_i] = np.interp(interp_time, time, mdot)
                if input["valve"]["flow"] == "filling":
                    self.mass_rate *= -1

            else:
                if input["valve"]["flow"] == "filling":
                    self.mass_rate[:] = -input["valve"]["mdot"]
                else:
                    self.mass_rate[:] = input["valve"]["mdot"]

        elif input["valve"]["type"] == "controlvalve":
            Cv = tp.cv_vs_time(
                self.Cv, 0, self.valve_time_constant, self.valve_characteristic
            )
            if input["valve"]["flow"] == "filling":
                Z = self.res_fluid.compressibility_factor()
                MW = self.MW
                k = self.res_fluid.cp0molar() / (self.res_fluid.cp0molar() - 8.314)
                self.mass_rate[0] = -tp.control_valve(
                    self.p_back, self.p0, self.T0, Z, MW, k, Cv
                )
            else:
                Z = self.fluid.compressibility_factor()
                MW = self.MW
                k = cpcv
                self.mass_rate[0] = tp.control_valve(
                    self.p0, self.p_back, self.T0, Z, MW, k, Cv
                )
        elif input["valve"]["type"] == "psv":
            if input["valve"]["flow"] == "filling":
                raise ValueError(
                    "Unsupported valve: ",
                    input["valve"]["type"],
                    " for vessel filling.",
                )
            self.mass_rate[0] = tp.relief_valve(
                self.p0,
                self.p_back,
                self.Pset,
                self.blowdown,
                cpcv,
                self.CD,
                self.T0,
                self.fluid.compressibility_factor(),
                self.MW,
                self.D_orifice**2 / 4 * math.pi,
            )

        self.time_array[0] = 0

        # setting heat transfer parameters
        T_profile, T_profile2 = 0, 0

        # Run actual integration by updating values by numerical integration/time stepping
        # Mass of fluid is calculated from previous time step mass and mass flow rate
        for i in tqdm(
            range(1, len(self.time_array)),
            desc="hyddown",
            disable=disable_pbar,
            total=len(self.time_array),
        ):
            self.time_array[i] = self.time_array[i - 1] + self.tstep
            self.mass_fluid[i] = (
                self.mass_fluid[i - 1] - self.mass_rate[i - 1] * self.tstep
            )

            self.rho[i] = self.mass_fluid[i] / self.vol

            # For the simple methods key variable is rho, with either H,S,U or T constant
            # updating T and p is convenient using coolprop.
            if self.method == "isenthalpic":
                self.fluid.update(CP.DmassHmass_INPUTS, self.rho[i], self.H_mass[i - 1])
                self.T_fluid[i] = self.fluid.T()
                self.P[i] = self.fluid.p()

            elif self.method == "isentropic":
                self.fluid.update(CP.DmassSmass_INPUTS, self.rho[i], self.S_mass[i - 1])
                self.T_fluid[i] = self.fluid.T()
                self.P[i] = self.fluid.p()

            elif self.method == "isothermal":
                self.fluid.update(CP.DmassT_INPUTS, self.rho[i], self.T0)
                self.T_fluid[i] = self.T0
                self.P[i] = self.fluid.p()

            elif self.method == "constantU":
                self.fluid.update(CP.DmassUmass_INPUTS, self.rho[i], self.U_mass[i - 1])
                self.T_fluid[i] = self.fluid.T()
                self.P[i] = self.fluid.p()

            # This is the **only** complicated part where we have the energy balance included
            elif self.method == "energybalance":
                # Finding the heat supplied/removed by heat transfer from/to outside
                if self.heat_method == "specified_h" or self.heat_method == "detailed":
                    if self.h_in == "calc":
                        if self.vessel_orientation == "horizontal":
                            L = self.diameter
                        else:
                            L = self.length
                        if input["valve"]["flow"] == "filling":
                            # T_film = (self.T_fluid[i - 1] + self.T_vessel[i - 1]) / 2
                            T_film = (
                                self.T_fluid[i - 1] + self.T_inner_wall[i - 1]
                            ) / 2
                            self.transport_fluid.update(
                                CP.PT_INPUTS, self.P[i - 1], T_film
                            )
                            hi = tp.h_inside_mixed(
                                L,
                                # self.T_vessel[i - 1],
                                self.T_inner_wall[i - i],
                                self.T_fluid[i - 1],
                                self.transport_fluid,
                                self.mass_rate[i - 1],
                                self.diameter,
                            )
                        else:
                            T_film = (
                                self.T_fluid[i - 1] + self.T_inner_wall[i - 1]
                            ) / 2
                            self.transport_fluid.update(
                                CP.PT_INPUTS, self.P[i - 1], T_film
                            )
                            hi = tp.h_inside(
                                L,
                                self.T_inner_wall[i - i],
                                self.T_fluid[i - 1],
                                self.transport_fluid,
                            )
                    else:
                        hi = self.h_in

                    self.h_inside[i] = hi

                    self.Q_inner[i] = (
                        self.surf_area_inner
                        * hi
                        * (self.T_inner_wall[i - 1] - self.T_fluid[i - 1])
                    )

                    self.Q_outer[i] = (
                        self.surf_area_outer
                        * self.h_out
                        * (self.Tamb - self.T_outer_wall[i - 1])
                    )
                    self.T_vessel[i] = self.T_vessel[i - 1] + (
                        self.Q_outer[i] - self.Q_inner[i]
                    ) * self.tstep / (
                        self.vessel_cp * self.vessel_density * self.vol_solid
                    )
                    if "thermal_conductivity" in self.input["vessel"].keys():
                        theta = 0.5  # Crank-Nicholson scheme
                        dt = self.tstep / 10
                        k, rho, cp = (
                            self.input["vessel"]["thermal_conductivity"],
                            self.vessel_density,
                            self.vessel_cp,
                        )
                        if (
                            "liner_thermal_conductivity"
                            not in self.input["vessel"].keys()
                        ):
                            nn = 11  # number of nodes
                            z = np.linspace(0, self.thickness, nn)
                            self.z = z
                            mesh = tm.Mesh(
                                z, tm.LinearElement
                            )  # Or `QuadraticElement` to
                            bc = [
                                {"q": self.Q_outer[i] / self.surf_area_outer},
                                {"q": -self.Q_inner[i] / self.surf_area_inner},
                            ]
                            cpeek = tm.isothermal_model(k, rho, cp)
                            domain = tm.Domain(mesh, [cpeek], bc)

                            if type(T_profile) == type(int()) and T_profile == 0:
                                domain.set_T(self.Tamb * np.ones(nn))
                            else:
                                domain.set_T(T_profile[-1, :])

                            solver = {"dt": dt, "t_end": self.tstep, "theta": theta}
                            t, T_profile = tm.solve_ht(domain, solver)

                            self.T_outer_wall[i] = T_profile[-1, 0]
                            self.T_inner_wall[i] = T_profile[-1, -1]
                        else:
                            k_liner = self.input["vessel"]["liner_thermal_conductivity"]
                            rho_liner = self.input["vessel"]["liner_density"]
                            cp_liner = self.input["vessel"]["liner_heat_capacity"]
                            liner = tm.isothermal_model(k_liner, rho_liner, cp_liner)
                            # shell = tm.isothermal_model(k, rho, cp)
                            shell = tm.isothermal_model(k, rho, cp)

                            # Domain and mesh for shell
                            thk = self.input["vessel"]["thickness"]  # thickness in m
                            nn = 11  # number of nodes
                            z_shell = np.linspace(0, thk, nn)  # node locations
                            # the bottom surface is located at z = 0.0

                            # # Domain and mesh for liner
                            thk = self.input["vessel"][
                                "liner_thickness"
                            ]  # thickness in m
                            # # nn = 11  # number of nodes
                            z_liner = np.linspace(-thk, 0, nn)  # node locations
                            # # the top surface is now located at z = 0.0
                            z2 = np.hstack((z_liner, z_shell[1:]))
                            self.z = z2
                            mesh2 = tm.Mesh(z2, tm.LinearElement)
                            for j, elem in enumerate(mesh2.elem):
                                if elem.nodes.mean() > 0.0:
                                    mesh2.subdomain[j] = 1
                            bc = [
                                {"q": -self.Q_inner[i] / self.surf_area_inner},
                                {"q": self.Q_outer[i] / self.surf_area_outer},
                            ]
                            domain2 = tm.Domain(mesh2, [liner, shell], bc)

                            if type(T_profile2) == type(int()) and T_profile2 == 0:
                                domain2.set_T(self.Tamb * np.ones(len(mesh2.nodes)))
                            else:
                                domain2.set_T(T_profile2[-1, :])
                            solver2 = {"dt": dt, "t_end": self.tstep, "theta": theta}
                            t_bonded, T_profile2 = tm.solve_ht(domain2, solver2)

                            self.T_outer_wall[i] = T_profile2[-1, -1]
                            self.T_inner_wall[i] = T_profile2[-1, 0]
                            self.T_bonded_wall[i] = T_profile2[-1, (nn - 1)]
                            self.temp_profile.append(T_profile2[-1, :])
                    else:
                        self.T_inner_wall[i] = self.T_vessel[i]
                        self.T_outer_wall[i] = self.T_vessel[i]

                elif self.heat_method == "s-b":
                    if self.vessel_orientation == "horizontal":
                        L = self.diameter
                    else:
                        L = self.length
                    hi = tp.h_inner(
                        L,
                        self.T_fluid[i - 1],
                        self.T_vessel[i - 1],
                        self.P[i - 1],
                        self.species,
                    )
                    self.h_inside[i] = hi
                    self.Q_inner[i] = (
                        self.surf_area_inner
                        * hi
                        * (self.T_vessel[i - 1] - self.T_fluid[i - 1])
                    )
                    self.Q_outer[i] = (
                        fire.sb_fire(self.T_vessel[i - 1], self.fire_type)
                        * self.surf_area_outer
                    )
                    self.T_vessel[i] = self.T_vessel[i - 1] + (
                        self.Q_outer[i] - self.Q_inner[i]
                    ) * self.tstep / (
                        self.vessel_cp * self.vessel_density * self.vol_solid
                    )

                elif self.heat_method == "specified_U":
                    self.Q_inner[i] = (
                        self.surf_area_outer
                        * self.Ufix
                        * (self.Tamb - self.T_fluid[i - 1])
                    )
                    self.T_vessel[i] = self.T_vessel[0]
                elif self.heat_method == "specified_Q":
                    self.Q_inner[i] = self.Qfix
                    self.T_vessel[i] = self.T_vessel[0]
                else:
                    self.Q_inner[i] = 0.0
                    self.T_vessel[i] = self.T_vessel[0]

                # NMOL = self.mass_fluid[i - 1] / self.MW
                # NMOL_ADD = (self.mass_fluid[i] - self.mass_fluid[i - 1]) / self.MW
                # New
                U_start = self.U_mass[i - 1] * self.mass_fluid[i - 1]

                # Smooting finction for very early times /numerical trick
                # Might not be necessary.
                x = 1 - math.exp(-1 * self.time_array[i]) ** 0.66

                # Finding the inlet/outlet enthalpy rate for the energy balance
                if input["valve"]["flow"] == "filling":
                    #h_in = self.fluid.hmass()
                    h_in = x * self.res_fluid.hmass() + (1 - x) * self.res_fluid.umass()

                else:
                    h_in = self.fluid.hmass()

                P2 = self.P[i - 1]
                if i > 1:
                    P1 = self.P[i - 2]
                else:
                    P1 = self.P[i - 1]

                U_end = (
                    U_start
                    - self.tstep * self.mass_rate[i - 1] * h_in
                    + self.tstep * self.Q_inner[i]
                )
                self.U_mass[i] = U_end / self.mass_fluid[i]
                P1, T1, self.U_res[i] = self.UDproblem(
                    U_end / self.mass_fluid[i],
                    self.rho[i],
                    self.P[i - 1],
                    self.T_fluid[i - 1],
                )

                self.P[i] = P1
                self.T_fluid[i] = T1
                self.fluid.update(CP.PT_INPUTS, self.P[i], self.T_fluid[i])

            else:
                raise NameError("Unknown calculation method: " + self.method)

            # Updating H,S,U states
            self.H_mass[i] = self.fluid.hmass()
            self.S_mass[i] = self.fluid.smass()
            self.U_mass[i] = self.fluid.umass()

            # print("Progress", int(i/(self.time_tot / self.tstep)*100),"%",end="\r")

            # Calculating vent temperature (adiabatic) only for discharge problem
            if self.input["valve"]["flow"] == "discharge":
                if "&" in self.species:
                    self.T_vent[i] = self.PHproblem(
                        self.H_mass[i], self.p_back, self.vent_fluid.T()
                    )
                else:
                    self.T_vent[i] = PropsSI(
                        "T", "H", self.H_mass[i], "P", self.p_back, self.species
                    )

            cpcv = self.fluid.cp0molar() / (self.fluid.cp0molar() - 8.314)

            # Finally updating the mass rate for the mass balance in the next time step
            if input["valve"]["type"] == "orifice":
                if input["valve"]["flow"] == "filling":
                    k = self.res_fluid.cp0molar() / (self.res_fluid.cp0molar() - 8.314)
                    self.mass_rate[i] = -tp.gas_release_rate(
                        self.p_back,
                        self.P[i],
                        self.res_fluid.rhomass(),
                        k,
                        self.CD,
                        self.D_orifice**2 / 4 * math.pi,
                    )
                else:
                    self.mass_rate[i] = tp.gas_release_rate(
                        self.P[i],
                        self.p_back,
                        self.rho[i],
                        cpcv,
                        self.CD,
                        self.D_orifice**2 / 4 * math.pi,
                    )
            elif input["valve"]["type"] == "controlvalve":
                Cv = tp.cv_vs_time(
                    self.Cv,
                    self.time_array[i],
                    self.valve_time_constant,
                    self.valve_characteristic,
                )
                if input["valve"]["flow"] == "filling":
                    Z = self.res_fluid.compressibility_factor()
                    MW = self.MW
                    k = self.res_fluid.cp0molar() / (self.res_fluid.cp0molar() - 8.314)
                    self.mass_rate[i] = -tp.control_valve(
                        self.p_back, self.P[i], self.T0, Z, MW, k, Cv
                    )
                else:
                    Z = self.fluid.compressibility_factor()
                    MW = self.MW
                    self.mass_rate[i] = tp.control_valve(
                        self.P[i], self.p_back, self.T_fluid[i], Z, MW, cpcv, Cv
                    )
            elif input["valve"]["type"] == "psv":
                self.mass_rate[i] = tp.relief_valve(
                    self.P[i],
                    self.p_back,
                    self.Pset,
                    self.blowdown,
                    cpcv,
                    self.CD,
                    self.T_fluid[i],
                    self.fluid.compressibility_factor(),
                    self.MW,
                    self.D_orifice**2 / 4 * math.pi,
                )
            if (
                "end_pressure" in self.input["valve"]
                and self.P[i] > self.input["valve"]["end_pressure"]
            ):
                massflow_stop_switch = 1
            if massflow_stop_switch:
                self.mass_rate[i] = 0
        self.isrun = True

    def get_dataframe(self):
        """
        Storing relevant results in pandas dataframe for e.g. export
        to csv, excel for archiving or post analysis
        """
        if self.isrun == True:
            df = pd.DataFrame(self.time_array, columns=["Time (s)"])

            df.insert(1, "Pressure (bar)", self.P / 1e5, True)
            df.insert(2, "Fluid temperature (oC)", self.T_fluid - 273.15, True)
            df.insert(3, "Wall temperature  (oC)", self.T_vessel - 273.15, True)
            df.insert(4, "Vent temperature  (oC)", self.T_vent - 273.15, True)
            df.insert(5, "Fluid enthalpy (J/kg)", self.H_mass, True)
            df.insert(6, "Fluid entropy (J/kg K)", self.S_mass, True)
            df.insert(7, "Fluid internal energy (J/kg)", self.U_mass, True)
            df.insert(8, "Discharge mass rate (kg/s)", self.mass_rate, True)
            df.insert(9, "Fluid mass (kg)", self.mass_fluid, True)
            df.insert(10, "Fluid density (kg/m3)", self.rho, True)
            df.insert(
                11, "Inner heat transfer coefficient (W/m2 K)", self.h_inside, True
            )
            df.insert(
                12,
                "Internal heat flux (W/m2)",
                self.Q_inner / self.surf_area_inner,
                True,
            )
            df.insert(
                13,
                "External heat flux (W/m2)",
                self.Q_outer / self.surf_area_outer,
                True,
            )
            df.insert(
                14, "Inner wall temperature  (oC)", self.T_inner_wall - 273.15, True
            )
            df.insert(
                13, "Outer wall temperature  (oC)", self.T_outer_wall - 273.15, True
            )
        return df

    def plot(self, filename=None, verbose=True):
        """
        Creating standard plots for the solved problem

        Parameters
        ----------
        filename : str
            Saving plots to filename if provideed (optional)
        verbose : bool
            Plotting on screen if True (optional)
        """
        import pylab as plt

        if filename != None:
            plt.figure(1, figsize=(12, 7), dpi=300)
        else:
            plt.figure(1, figsize=(8, 6))

        plt.subplot(221)
        plt.plot(self.time_array, self.T_fluid - 273.15, "b", label="Fluid")
        if "thermal_conductivity" not in self.input["vessel"].keys():
            plt.plot(self.time_array, self.T_vessel - 273.15, "g", label="Vessel")
        if "liner_thermal_conductivity" in self.input["vessel"].keys():
            plt.plot(
                self.time_array,
                self.T_bonded_wall - 273.15,
                "g",
                label="Liner/composite",
            )

        plt.plot(self.time_array, self.T_inner_wall - 273.15, "g--", label="Inner wall")
        plt.plot(self.time_array, self.T_outer_wall - 273.15, "g-.", label="Outer wall")
        if self.input["valve"]["flow"] == "discharge":
            plt.plot(self.time_array, self.T_vent - 273.15, "r", label="Vent")
        if "validation" in self.input:
            if "temperature" in self.input["validation"]:
                temp = self.input["validation"]["temperature"]
                if "gas_mean" in temp:
                    plt.plot(
                        np.asarray(temp["gas_mean"]["time"]),
                        np.asarray(temp["gas_mean"]["temp"]) - 273.15,
                        "b.",
                        label="Gas mean",
                    )
                if "gas_high" in temp:
                    plt.plot(
                        np.asarray(temp["gas_high"]["time"]),
                        np.asarray(temp["gas_high"]["temp"]) - 273.15,
                        "b-.",
                        label="Gas high",
                    )
                if "gas_low" in temp:
                    plt.plot(
                        np.asarray(temp["gas_low"]["time"]),
                        np.asarray(temp["gas_low"]["temp"]) - 273.15,
                        "b--",
                        label="Gas low",
                    )
                if "wall_mean" in temp:
                    plt.plot(
                        np.asarray(temp["wall_mean"]["time"]),
                        np.asarray(temp["wall_mean"]["temp"]) - 273.15,
                        "go",
                        label="Wall mean",
                    )
                if "wall_high" in temp:
                    plt.plot(
                        np.asarray(temp["wall_high"]["time"]),
                        np.asarray(temp["wall_high"]["temp"]) - 273.15,
                        "g-.",
                        label="Wall high",
                    )
                if "wall_inner" in temp:
                    plt.plot(
                        np.asarray(temp["wall_inner"]["time"]),
                        np.asarray(temp["wall_inner"]["temp"]) - 273.15,
                        "g+",
                        label="Inner wall",
                    )
                if "wall_low" in temp:
                    plt.plot(
                        np.asarray(temp["wall_low"]["time"]),
                        np.asarray(temp["wall_low"]["temp"]) - 273.15,
                        "g-.",
                        label="Wall high",
                    )
                if "wall_outer" in temp:
                    plt.plot(
                        np.asarray(temp["wall_outer"]["time"]),
                        np.asarray(temp["wall_outer"]["temp"]) - 273.15,
                        "gx",
                        label="Outer wall",
                    )

        plt.legend(loc="best")
        plt.xlabel("Time (seconds)")
        plt.ylabel("Temperature ($^\circ$C)")

        plt.subplot(222)
        plt.plot(self.time_array, self.P / 1e5, "b", label="Calculated")
        if "validation" in self.input:
            if "pressure" in self.input["validation"]:
                plt.plot(
                    np.asarray(self.input["validation"]["pressure"]["time"]),
                    self.input["validation"]["pressure"]["pres"],
                    "ko",
                    label="Experimental",
                )
        plt.legend(loc="best")
        plt.xlabel("Time (seconds)")
        plt.ylabel("Pressure (bar)")

        plt.subplot(223)
        plt.plot(self.time_array, self.H_mass, "b", label="H (J/kg)")
        plt.plot(self.time_array, self.U_mass, "g", label="U (J/kg)")
        plt.plot(self.time_array, self.S_mass * 100, "r", label="S*100 (J/kg K)")
        plt.legend(loc="best")
        plt.xlabel("Time (seconds)")
        plt.ylabel("Enthalpy/Internal Energy/Entropy")

        plt.subplot(224)
        plt.plot(self.time_array, self.mass_rate, "b", label="m_dot")
        plt.xlabel("Time (seconds)")
        plt.ylabel("Vent rate (kg/s)")

        if filename != None:
            plt.savefig(filename + "_main.png")

        if verbose:
            plt.show()

    def plot_envelope(self, filename=None, verbose=True):
        """
        Creating standard plots for the solved problem

        Parameters
        ----------
        filename : str
            Saving plots to filename if provideed (optional)
        verbose : bool
            Plotting on screen if True (optional)
        """
        import pylab as plt

        if filename != None:
            plt.figure(2, figsize=(12, 7), dpi=300)
        else:
            plt.figure(2, figsize=(8, 6))

        self.fluid.build_phase_envelope("None")
        PE = self.fluid.get_phase_envelope_data()

        plt.plot(PE.T, PE.p, "-", label="HEOS Phase Envelope", color="g")
        plt.plot(self.T_fluid, self.P, "-.", label="P/T fluid trajectory", color="b")
        plt.plot(self.T_fluid[0], self.P[0], "o", label="Start", color="b")
        plt.plot(self.T_fluid[-1], self.P[-1], ".", label="End", color="b")
        plt.xlabel("Temperature [K]")
        plt.ylabel("Pressure [Pa]")
        plt.legend(loc="best")
        plt.tight_layout()

        if filename != None:
            plt.savefig(filename + "_envelope.png")

        if verbose:
            plt.show()

    def plot_tprofile(self, filename=None, verbose=True):
        """
        Creating standard plots for the solved problem

        Parameters
        ----------
        filename : str
            Saving plots to filename if provideed (optional)
        verbose : bool
            Plotting on screen if True (optional)
        """

        # Add some checks if the profile has been constructed
        # return some
        import pylab as plt
        import numpy as np

        if filename != None:
            plt.figure(3, figsize=(8, 6))
        else:
            plt.figure(3, figsize=(8, 6))

        X, Y = np.meshgrid(self.z*1e3 , self.time_array[:-1])
        x0 = self.z[0]*1e3
        x1 = self.z[-1]*1e3
        y0 = self.time_array[0]
        y1 = self.time_array[-1]

        #plt.subplot(211)
        plt.contourf(Y,X, np.asarray(self.temp_profile), origin ='lower', levels=20)
        #plt.imshow(np.asarray(self.temp_profile).T, aspect = 'auto', extent=(y0,y1,x0,x1), origin='lower')

        plt.colorbar(label="Temperature (K)")
        plt.xlabel("Time (s)")
        plt.ylabel("z (mm)")
        # add title with descriptive text for z axis


        if filename != None:
                plt.savefig(filename + "_tprofile1.png", dpi=300)

        if filename != None:
            plt.figure(4, figsize=(8, 6))
        else:
            plt.figure(4, figsize=(8, 6))


        n = math.floor(len(self.time_array)/15)
        for i in range(len(self.time_array[::n])):
            plt.plot(self.temp_profile[::n][i], self.z*1e3,label=f"t = {int(self.time_array[::n][i])} s.")
        plt.legend(loc="best")
        plt.ylabel("z (mm)")
        plt.xlabel("Temperature (K)")
        plt.title("Temperature distribution")

        if filename != None:
            plt.savefig(filename + "_tprofile2.png", dpi=300)
        if verbose:
            plt.show()

    def __str__(self):
        return "HydDown vessel filling/depressurization class"

    def generate_report(self):
        """
        Generating a report summarising key features for the problem solved.
        Can be used for e.g. case studies, problem optimisation (external) etc.
        """
        report = {}

        report["start_time"] = self.time_array[0]
        report["end_time"] = self.time_array[-1]

        # Pressure
        report["max_pressure"] = max(self.P)
        report["time_max_pressure"] = self.time_array[np.argmax(self.P)]
        report["min_pressure"] = min(self.P)
        report["time_min_pressure"] = self.time_array[np.argmin(self.P)]

        # Temperatures
        report["max_fluid_temp"] = max(self.T_fluid)
        report["time_max_fluid_temp"] = self.time_array[np.argmax(self.T_fluid)]
        report["min_fluid_temp"] = min(self.T_fluid)
        report["time_min_fluid_temp"] = self.time_array[np.argmin(self.T_fluid)]

        report["max_wall_temp"] = max(self.T_vessel)
        report["time_max_wall_temp"] = self.time_array[np.argmax(self.T_vessel)]
        report["min_wall_temp"] = min(self.T_vessel)
        report["time_min_wall_temp"] = self.time_array[np.argmin(self.T_vessel)]

        report["max_inner_wall_temp"] = max(self.T_inner_wall)
        report["time_max_inner_wall_temp"] = self.time_array[
            np.argmax(self.T_inner_wall)
        ]
        report["min_inner_wall_temp"] = min(self.T_inner_wall)
        report["time_min_inner_wall_temp"] = self.time_array[
            np.argmin(self.T_inner_wall)
        ]

        report["max_outer_wall_temp"] = max(self.T_outer_wall)
        report["time_max_outer_wall_temp"] = self.time_array[
            np.argmax(self.T_outer_wall)
        ]
        report["min_outer_wall_temp"] = min(self.T_outer_wall)
        report["time_min_outer_wall_temp"] = self.time_array[
            np.argmin(self.T_outer_wall)
        ]

        # Mass flows and inventory
        report["max_mass_rate"] = max(self.mass_rate)
        report["initial_mass"] = self.mass_fluid[0]
        report["final_mass"] = self.mass_fluid[-1]
        report["volume"] = self.vol

        report["max_Q_inside"] = max(self.Q_inner)
        report["time_max_Q_inside"] = self.time_array[np.argmax(self.Q_inner)]
        report["max_heat_flux_inside"] = max(self.Q_inner / self.surf_area_inner)

        report["max_Q_outside"] = max(self.Q_outer)
        report["time_max_Q_outside"] = self.time_array[np.argmax(self.Q_outer)]
        report["max_heat_flux_outside"] = max(self.Q_outer / self.surf_area_outer)

        self.report = report



# HydDown hydrogen/other gas depressurisation
# Copyright (c) 2021 Anders Andreasen
# Published under an MIT license

from cerberus import Validator
from cerberus.errors import ValidationError


def validate_mandatory_ruleset(input):
    """
    Validate input

    Parameters
    ----------
    input : dict
        Structure holding input

    Return
    ----------
    retval : bool
        True for success, False for failure
    """
    schema_general = {
        "initial": {
            "required": True,
            "type": "dict",
            "allow_unknown": False,
            "schema": {
                "temperature": {"required": True, "type": "number"},
                "pressure": {"required": True, "type": "number"},
                "fluid": {"required": True, "type": "string"},
            },
        },
        "calculation": {
            "required": True,
            "type": "dict",
            "allow_unknown": False,
            "schema": {
                "type": {
                    "required": True,
                    "type": "string",
                    "allowed": [
                        "energybalance",
                        "isenthalpic",
                        "isentropic",
                        "isothermal",
                        "constant_U",
                    ],
                },
                "time_step": {"required": True, "type": "number", "min": 0.000001},
                "end_time": {"required": True, "type": "number", "min": 0},
            },
        },
        "vessel": {
            "required": True,
            "type": "dict",
            "allow_unknown": False,
            "schema": {
                "length": {"required": True, "type": "number"},
                "diameter": {"required": True, "type": "number"},
                "thickness": {"required": False, "type": "number", "min": 0.0},
                "heat_capacity": {"required": False, "type": "number", "min": 1},
                "thermal_conductivity": {
                    "required": False,
                    "type": "number",
                    "min": 0.0001,
                },
                "density": {"required": False, "type": "number", "min": 1},
                "liner_thickness": {"required": False, "type": "number", "min": 0.0},
                "liner_heat_capacity": {"required": False, "type": "number", "min": 1},
                "liner_thermal_conductivity": {
                    "required": False,
                    "type": "number",
                    "min": 0.0001,
                },
                "liner_density": {"required": False, "type": "number", "min": 1},
                "orientation": {
                    "required": False,
                    "type": "string",
                    "allowed": ["vertical", "horizontal"],
                },
            },
        },
        "valve": {
            "required": True,
            "type": "dict",
            "allow_unknown": False,
            "schema": {
                "type": {
                    "required": True,
                    "type": "string",
                    "allowed": ["orifice", "psv", "controlvalve", "mdot"],
                },
                "flow": {
                    "required": True,
                    "type": "string",
                    "allowed": ["discharge", "filling"],
                },
                "diameter": {"type": "number", "min": 0},
                "discharge_coef": {"type": "number", "min": 0},
                "set_pressure": {"type": "number", "min": 0},
                "end_pressure": {"type": "number", "min": 0},
                "blowdown": {"type": "number", "min": 0, "max": 1},
                "back_pressure": {"type": "number", "min": 0},
                "Cv": {"type": "number", "min": 0},
                "mdot": {"type": ["number", "list"]},
                "time": {"type": "list"},
                "time_constant": {"type": "number", "min": 0},
                "characteristic": {
                    "type": "string",
                    "allowed": ["linear", "eq", "fast"],
                },
            },
        },
        "heat_transfer": {
            "required": False,
            "type": "dict",
            "allow_unknown": False,
            "allowed": [
                "Q_fix",
                "h_inner",
                "h_outer",
                "temp_ambient",
                "type",
                "fire",
                "s-b",
                "D_throat",
            ],
            "schema": {
                "type": {
                    "type": "string",
                    "allowed": ["specified_Q", "specified_h", "specified_U", "s-b"],
                },
                "Q_fix": {"required": False, "type": "number"},
                "U_fix": {"required": False, "type": "number", "min": 0},
                "temp_ambient": {"required": False, "type": "number", "min": 0},
                "h_outer": {"required": False, "type": "number", "min": 0},
                "h_inner": {"required": False, "type": ["number", "string"]},
                "fire": {
                    "required": False,
                    "type": "string",
                    "allowed": [
                        "api_pool",
                        "api_jet",
                        "scandpower_pool",
                        "scandpower_jet",
                    ],
                },
                "D_throat": {"required": False, "type": "number", "min": 0},
            },
        },
        "validation": {
            "required": False,
            "type": "dict",
            "allow_unknown": False,
            "allowed": ["pressure", "temperature"],
            "schema": {
                "pressure": {
                    "type": "dict",
                    "required": False,
                    "contains": ["time", "pres"],
                    "schema": {
                        "pres": {
                            "required": False,
                            "type": "list",
                            "schema": {"type": "number"},
                        },
                        "time": {
                            "required": False,
                            "type": "list",
                            "schema": {"type": "number"},
                        },
                    },
                },
                "temperature": {
                    "type": "dict",
                    "required": False,
                    "allowed": [
                        "wall_high",
                        "wall_low",
                        "wall_mean",
                        "wall_inner",
                        "wall_outer",
                        "gas_high",
                        "gas_low",
                        "gas_mean",
                    ],
                    "schema": {
                        "gas_high": {
                            "required": False,
                            "type": "dict",
                            "contains": ["time", "temp"],
                            "schema": {
                                "temp": {
                                    "required": False,
                                    "type": "list",
                                    "schema": {"type": "number"},
                                },
                                "time": {
                                    "required": False,
                                    "type": "list",
                                    "schema": {"type": "number"},
                                },
                            },
                        },
                        "gas_low": {
                            "required": False,
                            "type": "dict",
                            "contains": ["time", "temp"],
                            "schema": {
                                "temp": {
                                    "required": False,
                                    "type": "list",
                                    "schema": {"type": "number"},
                                },
                                "time": {
                                    "required": False,
                                    "type": "list",
                                    "schema": {"type": "number"},
                                },
                            },
                        },
                        "gas_mean": {
                            "required": False,
                            "type": "dict",
                            "contains": ["time", "temp"],
                            "schema": {
                                "temp": {
                                    "required": False,
                                    "type": "list",
                                    "schema": {"type": "number"},
                                },
                                "time": {
                                    "required": False,
                                    "type": "list",
                                    "schema": {"type": "number"},
                                },
                            },
                        },
                        "wall_high": {
                            "required": False,
                            "type": "dict",
                            "contains": ["time", "temp"],
                            "schema": {
                                "temp": {
                                    "required": False,
                                    "type": "list",
                                    "schema": {"type": "number"},
                                },
                                "time": {
                                    "required": False,
                                    "type": "list",
                                    "schema": {"type": "number"},
                                },
                            },
                        },
                        "wall_mean": {
                            "required": False,
                            "type": "dict",
                            "contains": ["time", "temp"],
                            "schema": {
                                "temp": {
                                    "required": False,
                                    "type": "list",
                                    "schema": {"type": "number"},
                                },
                                "time": {
                                    "required": False,
                                    "type": "list",
                                    "schema": {"type": "number"},
                                },
                            },
                        },
                        "wall_low": {
                            "required": False,
                            "type": "dict",
                            "contains": ["time", "temp"],
                            "schema": {
                                "temp": {
                                    "required": False,
                                    "type": "list",
                                    "schema": {"type": "number"},
                                },
                                "time": {
                                    "required": False,
                                    "type": "list",
                                    "schema": {"type": "number"},
                                },
                            },
                        },
                        "wall_inner": {
                            "required": False,
                            "type": "dict",
                            "contains": ["time", "temp"],
                            "schema": {
                                "temp": {
                                    "required": False,
                                    "type": "list",
                                    "schema": {"type": "number"},
                                },
                                "time": {
                                    "required": False,
                                    "type": "list",
                                    "schema": {"type": "number"},
                                },
                            },
                        },
                        "wall_outer": {
                            "required": False,
                            "type": "dict",
                            "contains": ["time", "temp"],
                            "schema": {
                                "temp": {
                                    "required": False,
                                    "type": "list",
                                    "schema": {"type": "number"},
                                },
                                "time": {
                                    "required": False,
                                    "type": "list",
                                    "schema": {"type": "number"},
                                },
                            },
                        },
                    },
                },
            },
        },
    }

    v = Validator(schema_general)
    retval = v.validate(input)
    if v.errors:
        print(v.errors)

    return retval


def heat_transfer_validation(input):
    """
    Validate input['heat_transfer'] deeper than cerberus

    Parameters
    ----------
    input : dict
        Structure holding input


    Return
    ----------
        : bool
        True for success, False for failure
    """
    retval = True

    if input["calculation"]["type"] == "energybalance":
        if input["heat_transfer"]["type"] == "specified_h":
            schema_heattransfer = {
                "initial": {"required": True},
                "calculation": {"required": True},
                "validation": {"required": False},
                "valve": {"required": True},
                "vessel": {
                    "required": True,
                    "type": "dict",
                    "allow_unknown": False,
                    "schema": {
                        "length": {"required": True, "type": "number"},
                        "diameter": {"required": True, "type": "number"},
                        "thickness": {"required": True, "type": "number", "min": 0.0},
                        "heat_capacity": {"required": True, "type": "number", "min": 1},
                        "thermal_conductivity": {
                            "required": False,
                            "type": "number",
                            "min": 0.0001,
                        },
                        "density": {"required": True, "type": "number", "min": 1},
                        "liner_thickness": {
                            "required": False,
                            "type": "number",
                            "min": 0.0,
                        },
                        "liner_heat_capacity": {
                            "required": False,
                            "type": "number",
                            "min": 1,
                        },
                        "liner_thermal_conductivity": {
                            "required": False,
                            "type": "number",
                            "min": 0.0001,
                        },
                        "liner_density": {
                            "required": False,
                            "type": "number",
                            "min": 1,
                        },
                        "orientation": {
                            "required": True,
                            "type": "string",
                            "allowed": ["vertical", "horizontal"],
                        },
                    },
                },
                "heat_transfer": {
                    "required": True,
                    "type": "dict",
                    "allow_unknown": False,
                    "allowed": [
                        "h_inner",
                        "h_outer",
                        "temp_ambient",
                        "type",
                        "D_throat",
                    ],
                    "schema": {
                        "type": {"type": "string", "allowed": ["specified_h"]},
                        "temp_ambient": {"required": True, "type": "number", "min": 0},
                        "h_outer": {"required": True, "type": "number", "min": 0},
                        "h_inner": {"required": True, "type": ["number", "string"]},
                        "D_throat": {"required": False, "type": "number", "min": 0},
                    },
                },
            }

        elif input["heat_transfer"]["type"] == "specified_Q":
            schema_heattransfer = {
                "initial": {"required": True},
                "calculation": {"required": True},
                "validation": {"required": False},
                "valve": {"required": True},
                "vessel": {
                    "required": True,
                    "type": "dict",
                    "allow_unknown": False,
                    "schema": {
                        "length": {"required": True, "type": "number"},
                        "diameter": {"required": True, "type": "number"},
                        "thickness": {"required": False, "type": "number", "min": 0.0},
                        "heat_capacity": {
                            "required": False,
                            "type": "number",
                            "min": 1,
                        },
                        "density": {"required": False, "type": "number", "min": 1},
                        "orientation": {
                            "required": False,
                            "type": "string",
                            "allowed": ["vertical", "horizontal"],
                        },
                    },
                },
                "heat_transfer": {
                    "required": True,
                    "type": "dict",
                    "allow_unknown": False,
                    "allowed": ["Q_fix", "type"],
                    "schema": {
                        "type": {"type": "string", "allowed": ["specified_Q"]},
                        "Q_fix": {"required": False, "type": "number"},
                    },
                },
            }

        elif input["heat_transfer"]["type"] == "specified_U":
            schema_heattransfer = {
                "initial": {"required": True},
                "calculation": {"required": True},
                "validation": {"required": False},
                "valve": {"required": True},
                "vessel": {
                    "required": True,
                    "type": "dict",
                    "allow_unknown": False,
                    "schema": {
                        "length": {"required": True, "type": "number"},
                        "diameter": {"required": True, "type": "number"},
                        "thickness": {"required": False, "type": "number", "min": 0.0},
                        "heat_capacity": {
                            "required": False,
                            "type": "number",
                            "min": 1,
                        },
                        "density": {"required": False, "type": "number", "min": 1},
                        "orientation": {
                            "required": False,
                            "type": "string",
                            "allowed": ["vertical", "horizontal"],
                        },
                    },
                },
                "heat_transfer": {
                    "required": True,
                    "type": "dict",
                    "allow_unknown": False,
                    "allowed": ["U_fix", "type", "temp_ambient"],
                    "schema": {
                        "type": {"type": "string", "allowed": ["U_fix"]},
                        "U_fix": {"required": False, "type": "number", "min": 0.0},
                        "temp_ambient": {"required": True, "type": "number", "min": 0},
                    },
                },
            }

        elif input["heat_transfer"]["type"] == "s-b":
            schema_heattransfer = {
                "initial": {"required": True},
                "calculation": {"required": True},
                "validation": {"required": False},
                "valve": {"required": True},
                "vessel": {
                    "required": True,
                    "type": "dict",
                    "allow_unknown": False,
                    "schema": {
                        "length": {"required": True, "type": "number"},
                        "diameter": {"required": True, "type": "number"},
                        "thickness": {"required": True, "type": "number", "min": 0.0},
                        "heat_capacity": {"required": True, "type": "number", "min": 1},
                        "density": {"required": True, "type": "number", "min": 1},
                        "orientation": {
                            "required": True,
                            "type": "string",
                            "allowed": ["vertical", "horizontal"],
                        },
                    },
                },
                "heat_transfer": {
                    "required": True,
                    "type": "dict",
                    "allow_unknown": False,
                    "allowed": ["fire", "type"],
                    "schema": {
                        "type": {
                            "required": True,
                            "type": "string",
                            "allowed": ["s-b"],
                        },
                        "fire": {
                            "required": True,
                            "type": "string",
                            "allowed": [
                                "api_pool",
                                "api_jet",
                                "scandpower_pool",
                                "scandpower_jet",
                            ],
                        },
                    },
                },
            }
        v = Validator(schema_heattransfer)
        retval = v.validate(input)
        if v.errors:
            print(v.errors)

    return retval


def valve_validation(input):
    """
    Validate input['valve'] deeper than cerberus

    Parameters
    ----------
    input : dict
        Structure holding input


    Return
    ----------
        : bool
        True for success, False for failure
    """

    schema_psv = {
        "initial": {"required": True},
        "calculation": {"required": True},
        "validation": {"required": False},
        "vessel": {"required": True},
        "heat_transfer": {"required": False},
        "valve": {
            "required": True,
            "type": "dict",
            "allow_unknown": False,
            "schema": {
                "type": {
                    "required": True,
                    "type": "string",
                    "allowed": ["orifice", "psv", "controlvalve", "mdot"],
                },
                "flow": {
                    "required": True,
                    "type": "string",
                    "allowed": ["discharge", "filling"],
                },
                "diameter": {"required": True, "type": "number", "min": 0},
                "discharge_coef": {"required": True, "type": "number", "min": 0},
                "set_pressure": {"required": True, "type": "number", "min": 0},
                "end_pressure": {"type": "number", "min": 0},
                "blowdown": {"required": True, "type": "number", "min": 0, "max": 1},
                "back_pressure": {"required": True, "type": "number", "min": 0},
                "Cv": {"type": "number", "min": 0},
                "mdot": {"type": ["number", "list"]},
                "time": {"type": "list"},
            },
        },
    }

    schema_orifice = {
        "initial": {"required": True},
        "calculation": {"required": True},
        "validation": {"required": False},
        "vessel": {"required": True},
        "heat_transfer": {"required": False},
        "valve": {
            "required": True,
            "type": "dict",
            "allow_unknown": False,
            "schema": {
                "type": {
                    "required": True,
                    "type": "string",
                    "allowed": ["orifice", "psv", "controlvalve", "mdot"],
                },
                "flow": {
                    "required": True,
                    "type": "string",
                    "allowed": ["discharge", "filling"],
                },
                "diameter": {"required": True, "type": "number", "min": 0},
                "discharge_coef": {"required": True, "type": "number", "min": 0},
                "set_pressure": {"type": "number", "min": 0},
                "end_pressure": {"type": "number", "min": 0},
                "blowdown": {"type": "number", "min": 0, "max": 1},
                "back_pressure": {"required": True, "type": "number", "min": 0},
                "Cv": {"type": "number", "min": 0},
                "mdot": {"type": ["number", "list"]},
                "time": {"type": "list"},
            },
        },
    }

    schema_control_valve = {
        "initial": {"required": True},
        "calculation": {"required": True},
        "validation": {"required": False},
        "vessel": {"required": True},
        "heat_transfer": {"required": False},
        "valve": {
            "required": True,
            "type": "dict",
            "allow_unknown": False,
            "schema": {
                "type": {
                    "required": True,
                    "type": "string",
                    "allowed": ["orifice", "psv", "controlvalve", "mdot"],
                },
                "flow": {
                    "required": True,
                    "type": "string",
                    "allowed": ["discharge", "filling"],
                },
                "back_pressure": {"required": True, "type": "number", "min": 0},
                "Cv": {"required": True, "type": "number", "min": 0},
                "time_constant": {"type": "number", "min": 0},
                "characteristic": {
                    "type": "string",
                    "allowed": ["linear", "eq", "fast"],
                },
            },
        },
    }

    schema_mdot = {
        "initial": {"required": True},
        "calculation": {"required": True},
        "validation": {"required": False},
        "vessel": {"required": True},
        "heat_transfer": {"required": False},
        "valve": {
            "required": True,
            "type": "dict",
            "allow_unknown": False,
            "schema": {
                "type": {
                    "required": True,
                    "type": "string",
                    "allowed": ["orifice", "psv", "controlvalve", "mdot"],
                },
                "flow": {
                    "required": True,
                    "type": "string",
                    "allowed": ["discharge", "filling"],
                },
                "mdot": {"required": True, "type": ["number", "list"]},
                "time": {"type": ["number", "list"]},
                "back_pressure": {"required": True, "type": "number", "min": 0},
            },
        },
    }

    if input["valve"]["type"] == "psv":
        v = Validator(schema_psv)
        retval = v.validate(input)
        if v.errors:
            print(v.errors)
    elif input["valve"]["type"] == "orifice":
        v = Validator(schema_orifice)
        retval = v.validate(input)
        if v.errors:
            print(v.errors)
    elif input["valve"]["type"] == "controlvalve":
        v = Validator(schema_control_valve)
        retval = v.validate(input)
        if v.errors:
            print(v.errors)
    elif input["valve"]["type"] == "mdot":
        v = Validator(schema_mdot)
        retval = v.validate(input)
        if v.errors:
            print(v.errors)

    return retval


def validation(input):
    """
    Aggregate validation using cerberus

    Parameters
    ----------
    input : dict
        Structure holding input


    Return
    ----------
        : bool
        True for success, False for failure
    """
    return (
        validate_mandatory_ruleset(input)
        and valve_validation(input)
        and heat_transfer_validation(input)
    )



# HydDown hydrogen/other gas depressurisation
# Copyright (c) 2021 Anders Andreasen
# Published under an MIT license

import math
from CoolProp.CoolProp import PropsSI


def Gr(L, Tfluid, Tvessel, P, species):
    """
    Calculation of Grasshof number. See eq. 4.7-4 in
    C. J. Geankoplis Transport Processes and Unit Operations, International Edition,
    Prentice-Hall, 1993

    Parameters
    ----------
    L : float
        Vessel length
    Tfluid : float
        Temperature of the bulk fluid inventory
    Tvessel : float
        Temperature of the vessel wall (bulk)
    P : float
        Pressure of fluid inventory

    Returns
    ----------
    Gr : float
        Grasshof number
    """
    # Estimating the temperature at the fluid film interface
    T = (Tfluid + Tvessel) / 2
    beta = PropsSI("ISOBARIC_EXPANSION_COEFFICIENT", "T|gas", T, "P", P, species)
    nu = PropsSI("V", "T|gas", T, "P", P, 'HEOS::'+species.split('::')[1]) / PropsSI("D", "T|gas", T, "P", P, species)
    Gr = 9.81 * beta * abs(Tvessel - Tfluid) * L ** 3 / nu ** 2
    return Gr


def Pr(T, P, species):
    """
    Calculation of Prandtl number, eq. 4.5-6 in
    C. J. Geankoplis Transport Processes and Unit Operations, International Edition,
    Prentice-Hall, 1993

    Parameters
    ----------
    T : float
        Temperature of the fluid film interface
    P : float
        Pressure of fluid inventory

    Returns
    ----------
    Pr : float
        Prantdl number
    """
    C = PropsSI("C", "T|gas", T, "P", P, species)
    V = PropsSI("V", "T|gas", T, "P", P, 'HEOS::'+species.split('::')[1])
    L = PropsSI("L", "T|gas", T, "P", P, 'HEOS::'+species.split('::')[1])
    Pr = C * V / L

    return Pr


def Nu(Ra, Pr):
    """
    Calculation of Nusselt number for natural convection. See eq. 4.7-4  and Table 4.7-1 in
    C. J. Geankoplis Transport Processes and Unit Operations, International Edition,
    Prentice-Hall, 1993

    Parameters
    ----------
    Ra : float
        Raleigh number
    Pr : float
        Prandtl number

    Returns
    ----------
    Nu : float
        Nusselt numebr
    """
    if Ra >= 1e9:
        NNu = 0.13 * Ra ** 0.333
    elif Ra < 1e9 and Ra > 1e4:
        NNu = 0.59 * Ra ** 0.25
    else:
        NNu = 1.36 * Ra ** 0.20
    return NNu

def h_inside(L, Tvessel, Tfluid, fluid):
    """
    Calculation of internal natural convective heat transfer coefficient from Nusselt number
    and using the coolprop low level interface.

    Parameters
    ----------
    L : float
        Vessel length
    Tfluid : float
        Temperature of the bulk fluid inventory
    Tvessel : float
        Temperature of the vessel wall (bulk)
    fluid : obj
        Coolprop fluid object

    Returns
    ----------
    h_inner : float
        Heat transfer coefficient
    """
    cond = fluid.conductivity()
    visc = fluid.viscosity()
    cp = fluid.cpmass()
    Pr = cp * visc / cond
    beta = fluid.isobaric_expansion_coefficient()
    nu = visc / fluid.rhomass()
    Gr = 9.81 * beta * abs(Tvessel - Tfluid) * L ** 3 / nu ** 2
    Ra = Pr * Gr
    NNu = Nu(Ra, Pr)
    h_inner = NNu * cond / L
    return h_inner


def h_inner(L, Tfluid, Tvessel, P, species):
    """
    Calculation of internal natural convective heat transfer coefficient from Nusselt number
    and using the coolprop high level interface. Not currently in use.

    Parameters
    ----------
    L : float
        Vessel length
    Tfluid : float
        Temperature of the bulk fluid inventory
    Tvessel : float
        Temperature of the vessel wall (bulk)
    species : str
        Fluid definition string

    Returns
    ----------
    h_inner : float
        Heat transfer coefficient
    """
    NPr = Pr((Tfluid + Tvessel) / 2, P, species)
    NGr = Gr(L, Tfluid, Tvessel, P, species)
    NRa = NPr * NGr
    NNu = Nu(NRa, NPr)
    h_inner = NNu * PropsSI("L", "T|gas", (Tfluid + Tvessel) / 2, "P", P, 'HEOS::'+species.split('::')[1]) / L
    return h_inner


def h_inside_mixed(L, Tvessel, Tfluid, fluid, mdot, D):
    """
    Calculation of internal mixed natural/forced convective heat transfer coefficient from Nusselt number
    and using the coolprop low level interface.

    Parameters
    ----------
    L : float
        Vessel length
    Tfluid : float
        Temperature of the bulk fluid inventory
    Tvessel : float
        Temperature of the vessel wall (bulk)
    fluid : obj
        Coolprop fluid object
    mdot : float
        Mass flow
    D : float
        Characteristic diameter for Reynolds number estimation

    Returns
    ----------
    h_inner : float
        Heat transfer coefficient
    """
    cond = fluid.conductivity()
    visc = fluid.viscosity()
    cp = fluid.cpmass()
    Pr = cp * visc / cond

    T = (Tfluid + Tvessel) / 2
    beta = fluid.isobaric_expansion_coefficient()
    nu = visc / fluid.rhomass()
    Gr = 9.81 * beta * abs(Tvessel - Tfluid) * L ** 3 / nu ** 2
    Ra = Pr * Gr

    NNu_free = Nu(Ra,Pr)  # 0.13 * NRa**0.333
    Re = 4 * abs(mdot) / (visc * math.pi * D)
    NNu_forced = 0.56 * Re ** 0.67
    return (NNu_free + NNu_forced) * cond  / L


def h_inner_mixed(L, Tfluid, Tvessel, P, species, mdot, D):
    """
    Calculation of internal mixed (nutural/forced convective) heat transfer coefficient from Nusselt number
    and using the coolprop high level interface. Not currently in use.

    Parameters
    ----------
    L : float
        Vessel length
    Tfluid : float
        Temperature of the bulk fluid inventory
    Tvessel : float
        Temperature of the vessel wall (bulk)
    P : float
        Fluid pressure
    species : str
        Fluid definition string
    mdot : float
        Mass flow
    D : float
        Characteristic diameter for Reynolds number estimation


    Returns
    ----------
    h_inner : float
        Heat transfer coefficient
    """
    NPr = Pr((Tfluid + Tvessel) / 2, P, species)
    NGr = Gr(L, Tfluid, Tvessel, P, species)
    NRa = NPr * NGr
    NNu_free = Nu(NRa,NPr)  # 0.13 * NRa**0.333
    Re = 4 * abs(mdot) / (PropsSI("V", "T|gas", Tfluid, "P", P, 'HEOS::'+species.split('::')[1]) * math.pi * D)
    NNu_forced = 0.56 * Re ** 0.67
    return (
        (NNu_free + NNu_forced)
        * PropsSI("L", "T|gas", (Tfluid + Tvessel) / 2, "P", P, 'HEOS::'+species.split('::')[1])
        / L
    )


def gas_release_rate(P1, P2, rho, k, CD, area):
    """
    Gas massflow (kg/s) trough a hole at critical (sonic) or subcritical
    flow conditions. The formula is based on Yellow Book equation 2.22.

    Methods for the calculation of physical effects, CPR 14E, van den Bosch and Weterings (Eds.), 1996

    Parameters
    ----------
    P1 : float
        Upstream pressure
    P2 : float
        Downstream pressure
    rho : float
        Fluid density
    k : float
        Ideal gas k (Cp/Cv)
    CD : float
        Coefficient of discharge
    are : float
        Orifice area

    Returns
    ----------
        : float
        Gas release rate / mass flow of discharge
    """
    if P1 > P2:
        if P1 / P2 > ((k + 1) / 2) ** ((k) / (k - 1)):
            flow_coef = 1
        else:
            flow_coef = (
                2
                / (k - 1)
                * (((k + 1) / 2) ** ((k + 1) / (k - 1)))
                * ((P2 / P1) ** (2 / k))
                * (1 - (P2 / P1) ** ((k - 1) / k))
            )

        return (
            math.sqrt(flow_coef)
            * CD
            * area
            * math.sqrt(rho * P1 * k * (2 / (k + 1)) ** ((k + 1) / (k - 1)))
        )
    else:
        return 0


def relief_valve(P1, Pback, Pset, blowdown, k, CD, T1, Z, MW, area):
    """
    Pop action relief valve model including hysteresis.
    The pressure shall rise above P_set to open and
    decrease below P_reseat (P_set*(1-blowdown)) to close

    Parameters
    ----------
    P1 : float
        Upstream pressure
    Pback : float
        Downstream / backpressure
    Pset : float
        Set pressure of the PSV / relief valve
    blowdown : float
        The percentage of the set pressure at which the valve reseats
    k : float
        Ideal gas k (Cp/Cv)
    CD : float
        Coefficient of discharge
    T1 : float
        Upstream temperature
    Z : float
        Compressibility
    MW : float
        Molecular weight of the gas relieved
    area : float
        PSV orifice area

    Returns
    ----------
        : float
        Relief rate / mass flow
    """

    global psv_state
    if P1 > Pset:
        eff_area = area
        psv_state = "open"
    elif P1 < Pset * (1 - blowdown):
        eff_area = 0
        psv_state = "closed"
    else:
        if psv_state == "open":
            eff_area = area
        elif psv_state == "closed":
            eff_area = 0
        else:
            raise ValueError("Unknown PSV open/close state.")

    if eff_area > 0:
        return api_psv_release_rate(P1, Pback, k, CD, T1, Z, MW, area)
    else:
        return 0.0


def api_psv_release_rate(P1, Pback, k, CD, T1, Z, MW, area):
    """
    PSV vapour relief rate calculated according to API 520 Part I 2014
    Eq. 5, 9, 15, 18

    Parameters
    ----------
    P1 : float
        Upstream pressure
    Pback : float
        Downstream / backpressure
    k : float
        Ideal gas k (Cp/Cv)
    CD : float
        Coefficient of discharge
    T1 : float
        Upstream temperature
    Z : float
        Compressibility
    MW : float
        Molecular weight of the gas relieved
    area : float
        PSV orifice area

    Returns
    ----------
        : float
        Relief rate / mass flow
    """


    P1 = P1 / 1000
    Pback = Pback / 1000
    area = area * 1e6
    MW = MW * 1000
    C = 0.03948 * (k * (2 / (k + 1)) ** ((k + 1) / (k - 1))) ** 0.5
    if P1 / Pback > ((k + 1) / 2) ** ((k) / (k - 1)):
        w = CD * area * C * P1 / math.sqrt(T1 * Z / MW)
    else:
        r = Pback / P1
        f2 = ((k / (k - 1)) * r ** (2 / k) * (1 - r**((k - 1) / k)) / (1 - r))**0.5
        print(f2)
        w = CD * area * f2 / (T1 * Z / (MW * P1 * (P1 - Pback)))**0.5 / 17.9
    return w/3600

def cv_vs_time(Cv_max,t,time_constant=0,characteristic="linear"):
    """
    Control valve flow coefficient vs time / actuator postion
    assuming a linear rate of actuator for the three archetypes of
    characteristics: linear, equal percentage and fast/quick opening.

    Parameters
    ----------
    Cv_max : float
        Valve flow coefficient at full open position
    t : float
        Time
    time_constant : float (optional)
        The time required for the actuator to fully open.
        Default to instant open
    characteristic : string (optional)
        Valve characteristic
        Default to linear.
    """

    if time_constant == 0:
        return Cv_max
    else:
        if characteristic=="linear":
            return Cv_max * min(t/time_constant,1)
        elif characteristic=="eq":
            # https://www.spiraxsarco.com/learn-about-steam/control-hardware-electric-pneumatic-actuation/control-valve-characteristics
            tau=50
            travel=min(t/time_constant,1)
            return Cv_max * math.exp( travel * math.log(tau)) / tau
        elif characteristic=="fast":
            # square root function used
            return Cv_max * min(t/time_constant,1)**(0.5)
        else:
            return Cv_max

def control_valve(P1, P2, T, Z, MW, gamma, Cv, xT=0.75, FP=1):
    """
    Flow calculated from ANSI/ISA control valve equations for single phase gas flow.
    Equation 19 pp. 132 in
    Control Valves / Guy Borden, editor; Paul Friedmann, style editor

    Parameters
    ----------
    P1 : float
        Upstream pressure
    P2 : float
        Downstream / backpressure
    T : float
        Upstream temperature
    Z : float
        Upstream compressibility
    MW : float
        Molecular weight of the gas relieved
    gamma : float
        Upstream Ideal gas k (Cp/Cv)
    Cv : float
        Valve coefficient
    xT : float
        Value of xT for valve fitting assembly, default value
    FP : float
        Piping geometry factor

    Returns
    ----------
        : float
        Mass flow
    """

    P1 = P1 / 1e5
    P2 = P2 / 1e5
    MW = MW * 1000
    N8 = 94.8
    Fk = gamma / 1.4
    x = (P1 - P2) / P1
    if x < 0:
        x = 0
    Y = 1.0 - min(x, Fk * xT) / (3.0 * Fk * xT)
    mass_flow = N8 * FP * Cv * P1 * Y * (MW * min(x, xT * Fk) / T / Z) ** 0.5
    return mass_flow / 3600  # kg/s



# HydDown hydrogen/other gas depressurisation
# Copyright (c) 2021 Anders Andreasen
# Published under an MIT license

import streamlit as st
import pandas as pd
from PIL import Image
import base64
import matplotlib.pyplot as plt

try:
    from hyddown import HydDown
except:
    import sys
    import os

    hyddown_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), "..", "src")
    sys.path.append(os.path.abspath(hyddown_path))
    from hyddown import HydDown


def get_table_download_link(df, filename):
    """
    Generates a link allowing the data in a given panda dataframe to be downloaded
    in:  dataframe
    out: href string
    """
    csv = df.to_csv(index=False)
    b64 = base64.b64encode(
        csv.encode()
    ).decode()  # some strings <-> bytes conversions necessary here
    filename = filename + ".csv"
    return f'<a href="data:application/octet-stream;base64,{b64}" download={filename}>Download csv file</a>'


def read_input():
    sideb = st.sidebar

    with sideb:
        try:
            image_path = os.path.join(
                os.path.abspath(os.path.dirname(__file__)),
                "..",
                "docs",
                "img",
                "Sketch.png",
            )
            icon = Image.open(image_path)
            st.image(icon, use_column_width=True, caption="HydDown")
        except:
            pass

        with st.form(key="my_form"):
            submit_button = st.form_submit_button(label="Run calculation")
            heattran = st.checkbox("Include heat transfer", value=True)
            c1, c2 = st.columns(2)

            with c2:
                length = st.text_input("Vessel length (m):", 0.463)
                diam = st.text_input("Vessel diam (m):", 0.254)
                thk = st.text_input("Vessel thichness (m):", 0.016)
                orientation = st.selectbox(
                    "Vessel orientation", ("horizontal", "vertical")
                )
                orifice_diam = st.text_input("Orifice diam (mm):", 0.40)
                orifice_diam = float(orifice_diam) / 1000
                tstep = st.text_input("Time step (s):", 1.0)

            with c1:
                pres = st.text_input("Initial pressure (bar):", 50.0)
                pres = float(pres) * 1e5

                back_pressure = st.text_input("Fill/back pres. (bar):", 240)
                back_pressure = float(back_pressure) * 1e5

                #fluid = st.selectbox('Select fluid', ('H2', 'He', 'N2', 'air', 'CH4', 'O2'))
                fluid = st.selectbox('Select fluid', ('H2', 'NG', 'NG1', 'He', 'N2', 'air', 'CH4','O2'))
                if fluid == 'NG':
                    fluid = "Methane[0.89571]&Ethane[5.6739e-02]&Propane[2.30395e-02]&Butane[1.03E-02]&Pentane[2.67E-03]&CO2[0.84e-02]&N2[0.3080e-2]"
                if fluid == 'NG1':
                    fluid = "Methane[0.860231]&Ethane[0.078217]&Propane[0.033786]&Butane[9.210E-03]&Pentane[2.573E-03]&Hexane[3.560E-04]&CO2[1.206E-02]&N2[3.701E-03]"
    
                mode = st.selectbox('Select mode', ('filling', 'discharge'))
                temp = st.text_input("Initial temp. (C):", 25)
                temp = float(temp) + 273.15
                end_time = st.text_input("End time (s):", 240)

            density = st.text_input("Vessel material density (kg/m3):", 7740)
            density = float(density)

            cp = st.text_input("Vessel material heat capacity (J/kg K):", 470)
            cp = float(cp)

    input = {}
    input["calculation"] = {}
    input["vessel"] = {}
    input["initial"] = {}
    input["valve"] = {}
    input["heat_transfer"] = {}

    input["calculation"]["type"] = "energybalance"
    input["calculation"]["time_step"] = float(tstep)
    input["calculation"]["end_time"] = float(end_time)

    input["vessel"]["length"] = float(length)
    input["vessel"]["diameter"] = float(diam)
    input["vessel"]["heat_capacity"] = cp
    input["vessel"]["density"] = density
    input["vessel"]["orientation"] = orientation
    input["vessel"]["thickness"] = float(thk)

    input["initial"]["pressure"] = pres
    input["initial"]["temperature"] = temp
    input["initial"]["fluid"] = fluid
    input["valve"]["flow"] = mode
    input["valve"]["type"] = "orifice"
    input["valve"]["diameter"] = float(orifice_diam)
    input["valve"]["discharge_coef"] = 0.84
    input["valve"]["back_pressure"] = back_pressure
    # input['valve']['end_pressure']=end_pressure

    input["heat_transfer"]["type"] = "specified_h"
    input["heat_transfer"]["temp_ambient"] = 298
    input["heat_transfer"]["h_outer"] = 5
    if heattran is True:
        input["heat_transfer"]["h_inner"] = "calc"
    else:
        input["heat_transfer"]["h_inner"] = 0.0
    input["heat_transfer"]["D_throat"] = float(diam)
    return input


if __name__ == "__main__":
    st.set_page_config(layout="wide")

    input = read_input()
    hdown = HydDown(input)

    with st.spinner("Calculating, please wait...."):
        hdown.run(disable_pbar=True)

    st.title("HydDown rigorous demo")
    st.subheader(r"https://github.com/andr1976/HydDown")
    my_expander = st.expander("Description")

    my_expander.write(
        "Real gas vessel pressurisation/depressurisation with heat transfer from gas to vessel and ambient and vice versa. Orifice size (Cd = 0.84) is specified for desired pressurisation/depressurisation rate."
    )
    my_expander.write(
        "For more information about the calculations and validation of the code please refer to the [manual](https://github.com/andr1976/HydDown/raw/main/docs/MANUAL.pdf)"
    )

    df = hdown.get_dataframe()
    file_name = st.text_input("Filename for saving data:", "saved_data")

    st.markdown(get_table_download_link(df, file_name), unsafe_allow_html=True)

    col1, col2 = st.columns(2)

    if input["valve"]["flow"] == "discharge":
        temp_data = pd.DataFrame(
            {
                "Time (s)": hdown.time_array,
                "Fluid temperature (C)": hdown.T_fluid - 273.15,
                "Wall temperature (C)": hdown.T_vessel - 273.15,
                "Vent temperature (C)": hdown.T_vent - 273.15,
            }
        )
    else:
        temp_data = pd.DataFrame(
            {
                "Time (s)": hdown.time_array,
                "Fluid temperature (C)": hdown.T_fluid - 273.15,
                "Wall temperature (C)": hdown.T_vessel - 273.15,
            }
        )

    pres_data = pd.DataFrame(
        {"Time (s)": hdown.time_array, "Pressure (bar)": hdown.P / 1e5}
    )

    fig, ax = plt.subplots(figsize=(5, 2))

    ax.plot(
        pres_data["Time (s)"],
        pres_data["Pressure (bar)"],
        "k",
    )
    ax.set_xlabel("Time (s)")
    ax.set_ylabel("Pressure (bar)")
    col1.pyplot(fig)

    fig, ax = plt.subplots(figsize=(5, 2))

    ax.plot(
        temp_data["Time (s)"], temp_data["Fluid temperature (C)"], "k", label="Fluid"
    )
    ax.plot(
        temp_data["Time (s)"], temp_data["Wall temperature (C)"], "k--", label="Wall"
    )
    if input["valve"]["flow"] == "discharge":
        ax.plot(
            temp_data["Time (s)"],
            temp_data["Vent temperature (C)"],
            "k-.",
            label="Vent",
        )

    ax.set_xlabel("Time (s)")
    ax.set_ylabel("Temperature ($^\circ$C)")
    ax.legend(loc="best")
    col2.pyplot(fig)

    mdot_data = pd.DataFrame(
        {"Time (s)": hdown.time_array, "Mass rate (kg/s)": hdown.mass_rate}
    )
    mass_data = pd.DataFrame(
        {"Time (s)": hdown.time_array, "Fluid inventory (kg)": hdown.mass_fluid}
    )

    fig, ax = plt.subplots(figsize=(5, 2))
    ax.plot(
        mdot_data["Time (s)"],
        mdot_data["Mass rate (kg/s)"],
        "k",
    )
    ax.set_xlabel("Time (s)")
    ax.set_ylabel("Mass rate (kg/s)")
    col1.pyplot(fig)

    fig, ax = plt.subplots(figsize=(5, 2))
    ax.plot(
        mass_data["Time (s)"],
        mass_data["Fluid inventory (kg)"],
        "k",
    )
    ax.set_xlabel("Time (s)")
    ax.set_ylabel("Fluid inventory (kg)")
    col2.pyplot(fig)
